#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <fstream>
#include <sstream>
#include <iostream>
#include <iomanip>
#include <errno.h>

#ifdef HAVE_SYS_PRCTL_H
#include <sys/prctl.h>
#endif

#ifdef HAVE_SYSTEMD_SD_DAEMON_H
#include <systemd/sd-daemon.h>
#endif

#include "arsockserver.hh"
#include "airdata.h"
#include "icaofpl.h"
#include "wmm.h"

SocketServer::Client::Client(SocketServer *server, const Glib::RefPtr<Gio::SocketConnection>& connection, bool log)
	: m_server(server), m_connection(connection), m_refcount(1), m_closeprotect(0), m_firstreply(true), m_log(log)
{
	if (true)
		std::cout << "client create " << (unsigned long long)this << " pid " << getpid() << std::endl;
	if (!m_connection || !m_server) {
		m_server = 0;
		m_connection.reset();
		return;
	}
	m_connection->get_socket()->property_blocking() = false;
	m_connin = Glib::signal_io().connect(sigc::mem_fun(*this, &Client::on_input), m_connection->get_socket()->get_fd(), Glib::IO_IN | Glib::IO_ERR | Glib::IO_HUP);
	m_server->add_client(get_ptr());
	if (true)
		std::cout << "client connect " << (unsigned long long)this << " pid " << getpid() << " fd " << m_connection->get_socket()->get_fd() << std::endl;
}

SocketServer::Client::~Client()
{
	close();
	if (true)
		std::cout << "client destroy " << (unsigned long long)this << " pid " << getpid() << std::endl;
}

void SocketServer::Client::reference(void) const
{
        g_atomic_int_inc(&m_refcount);
	if (false)
		std::cout << "client " << (unsigned long long)this << " pid " << getpid() << " reference " << m_refcount << std::endl;
}

void SocketServer::Client::unreference(void) const
{
	bool nz(!g_atomic_int_dec_and_test(&m_refcount));
	if (false)
		std::cout << "client " << (unsigned long long)this << " pid " << getpid() << " unreference " << m_refcount << std::endl;
        if (nz)
                return;
        delete this;
}

void SocketServer::Client::close(void)
{
	m_connin.disconnect();
	m_connout.disconnect();
	m_sendqueue.clear();
	if (m_longpoll)
		m_longpoll->unset_longpoll(get_ptr());
	m_longpoll.reset();
	m_logfilter.clear();
	m_logdiscard.clear();
	if (!m_connection)
		return;
	if (true)
		std::cout << "client " << (unsigned long long)this << " pid " << getpid() << " close fd " << m_connection->get_socket()->get_fd() << std::endl;
	m_connection->get_socket()->close();
	m_connection.reset();
	if (m_closeprotect == 1)
		m_closeprotect = 2;
	else if (m_server)
		m_server->remove_client(get_ptr());
}

void SocketServer::Client::send(const Json::Value& v)
{
	if (!m_connection)
		return;
	m_sendqueue.push_back(v);
	m_connout.disconnect();
	m_connout = Glib::signal_io().connect(sigc::mem_fun(*this, &Client::on_output), m_connection->get_socket()->get_fd(), Glib::IO_OUT | Glib::IO_ERR);
	m_firstreply = false;
}

bool SocketServer::Client::on_output(Glib::IOCondition iocond)
{
	m_connout.disconnect();
	if (m_sendqueue.empty())
		return true;
	Json::FastWriter writer;
	std::string msg(writer.write(*m_sendqueue.begin()));
	try {
		gssize r(m_connection->get_socket()->send(msg.c_str(), msg.size()));
		if (r == -1) {
			if (true)
				std::cerr << "client " << (unsigned long long)this << " pid " << getpid() << " socket output return value: " << r << std::endl;
			close();
			return true;
		}
		m_sendqueue.erase(m_sendqueue.begin());
		if (m_log)
			std::cout << "client " << (unsigned long long)this << " pid " << getpid() << " tx " << msg << std::endl;
	} catch (const Gio::Error& error) {
		if (error.code() == Gio::Error::WOULD_BLOCK) {
			if (true)
				std::cerr << "client " << (unsigned long long)this << " pid " << getpid() << " socket output would block" << std::endl;
		} else {
			if (true)
				std::cerr << "client " << (unsigned long long)this << " pid " << getpid() << " socket output error: " << error.what() << std::endl;
			close();
			return true;
		}
	}
	if (!m_sendqueue.empty())
		m_connout = Glib::signal_io().connect(sigc::mem_fun(*this, &Client::on_output), m_connection->get_socket()->get_fd(), Glib::IO_OUT | Glib::IO_ERR);
	return true;
}

bool SocketServer::Client::on_input(Glib::IOCondition iocond)
{
	if (!m_connection) {
		m_connin.disconnect();
		return true;
	}
	if (!m_server)
		return false;
	if (m_server->is_sockclient()) {
		if (true)
			std::cerr << "client " << (unsigned long long)this << " pid " << getpid() << " receive in socket client" << std::endl;
		close();
		return false;
	}
	Json::Value root;
	Json::Value reply;
	if (m_firstreply) {
		reply["version"] = VERSION;
		reply["provider"] = "cfmu";
	}
	sockcli_timestamp(reply);
	try {
		char buf[16384];
 		gssize r(m_connection->get_socket()->receive(buf, sizeof(buf)));
		if (r <= 0) {
			if (true)
				std::cerr << "client " << (unsigned long long)this << " pid " << getpid() << " socket input return value: " << r << std::endl;
			close();
			return true;
		}
		if (r >= sizeof(buf)) {
			if (true)
				std::cerr << "client " << (unsigned long long)this << " pid " << getpid() << " socket input return value: " << r << std::endl;
			reply["error"] = std::string("invalid message size");
			send(reply);
			return true;
		}
		{
			Json::Reader reader;
			if (!reader.parse(buf, buf + r, root, false)) {
				if (true)
					std::cerr << "client " << (unsigned long long)this << " pid " << getpid() << " cannot parse JSON: " << std::string(buf, buf + r);
				reply["error"] = std::string("cannot parse JSON");
				send(reply);
				return true;
			}
		}
		if (m_log)
			std::cout << "client " << (unsigned long long)this  << " pid " << getpid()<< " rx " << std::string(buf, buf + r) << std::endl;
	} catch (const Gio::Error& error) {
		if (error.code() == Gio::Error::WOULD_BLOCK) {
			if (true)
				std::cerr << "client " << (unsigned long long)this << " pid " << getpid() << " socket input would block" << std::endl;
			return true;
		} else {
			if (true)
				std::cerr << "client " << (unsigned long long)this << " pid " << getpid() << " socket input error: " << error.what() << std::endl;
			close();
			return true;
		}
	}
	if (m_longpoll)
		m_longpoll->unset_longpoll(get_ptr());
	m_longpoll.reset();
	m_logfilter.clear();
	m_logdiscard.clear();
	if (root.isMember("logfilter")) {
		const Json::Value& filters(root["logfilter"]);
		if (filters.isArray()) {
			for (Json::ArrayIndex i = 0; i < filters.size(); ++i) {
				const Json::Value& filter(filters[i]);
				if (!filter.isString())
					continue;
				std::string s(filter.asString());
				if (s.empty())
					continue;
				m_logfilter.insert(s);
			}
		}
	}
	if (root.isMember("logdiscard")) {
		const Json::Value& filters(root["logdiscard"]);
		if (filters.isArray()) {
			for (Json::ArrayIndex i = 0; i < filters.size(); ++i) {
				const Json::Value& filter(filters[i]);
				if (!filter.isString())
					continue;
				std::string s(filter.asString());
				if (s.empty())
					continue;
				m_logdiscard.insert(s);
			}
		}
	}
	bool quit;
	{
		std::string sess;
		if (root.isMember("session"))
                        sess = root.get("session", "").asString();
		if (sess.empty()) {
			if (root.isMember("cmds")) {
				const Json::Value& cmds(root["cmds"]);
				if (cmds.isObject()) {
					Json::Value& replycmds(reply["cmds"]);
					Json::Value cmdout(m_server->servercmd(cmds));
					if (cmdout.isNull()) {
						cmdout["error"] = std::string("no session");
						if (cmds.isMember("cmdseq"))
							cmdout["cmdseq"] = cmds["cmdseq"];
						if (cmds.isMember("cmdname"))
							cmdout["cmdname"] = cmds["cmdname"];
					}
					sockcli_timestamp(cmdout);
					replycmds.append(cmdout);
				} else if (cmds.isArray()) {
					Json::Value& replycmds(reply["cmds"]);
					for (Json::ArrayIndex i = 0; i < cmds.size(); ++i) {
						const Json::Value& cmd(cmds[i]);
						if (!cmd.isObject())
							continue;
						Json::Value cmdout(m_server->servercmd(cmd));
						if (cmdout.isNull()) {
							cmdout["error"] = std::string("no session");
							if (cmd.isMember("cmdseq"))
								cmdout["cmdseq"] = cmd["cmdseq"];
							if (cmd.isMember("cmdname"))
								cmdout["cmdname"] = cmd["cmdname"];
						}
						sockcli_timestamp(cmdout);
						replycmds.append(cmdout);
					}
				}
			}
			send(reply);
			return true;
		}
		reply["session"] = sess;
		if (!m_server) {
			reply["error"] = std::string("no server");
			send(reply);
			return true;
		}
		quit = root.isMember("quit");
		if (quit) {
			const Json::Value& q(root["quit"]);
			quit = q.isBool() && q.asBool();
		}
		m_closeprotect = 1;
		m_longpoll = m_server->find_router(sess, !quit);
		if (m_closeprotect == 2) {
			m_closeprotect = 0;
			if (m_server)
				m_server->remove_client(get_ptr());
			return false;
		}
		m_closeprotect = 0;
	}
	if (!m_longpoll) {
		if (quit) {
			reply["quit"] = false;
		} else {
			reply["error"] = std::string("server busy");
		}
		send(reply);
		return true;
	}
       	if (root.isMember("cmds")) {
		const Json::Value& cmds(root["cmds"]);
		if (cmds.isObject()) {
			Json::Value cmdout(m_server->servercmd(cmds));
			if (!cmdout.isNull())
				send(cmdout);
			else
				m_longpoll->send(cmds);
		} else if (cmds.isArray()) {
			for (Json::ArrayIndex i = 0; i < cmds.size(); ++i) {
				const Json::Value& cmd(cmds[i]);
				if (!cmd.isObject())
					continue;
				Json::Value cmdout(m_server->servercmd(cmd));
				if (!cmdout.isNull())
					send(cmdout);
				else
					m_longpoll->send(cmd);
			}
		}
	}
	if (!quit && root.isMember("noreply")) {
		const Json::Value& noreply(root["noreply"]);
		if (noreply.isBool() && noreply.asBool()) {
			m_longpoll.reset();
			send(reply);
			return true;
		}
	}
	if (quit)
		m_longpoll->close("quit");
	m_longpoll->receive(reply, m_logfilter, m_logdiscard);
	Json::Value& replycmds(reply["cmds"]);
	if (quit) {
		reply["quit"] = true;
		unsigned int rxsz(m_longpoll->get_recvqueue_size());
		reply["rxqueue"] = rxsz;
		if (!rxsz || (root.isMember("clearrxqueue") && root["clearrxqueue"].isBool() && root["clearrxqueue"].asBool()))
			m_longpoll->zap();
		send(reply);
		return true;
	}
	if (replycmds.size()) {
		send(reply);
		return true;
	}
	m_longpoll->set_longpoll(get_ptr());
	return true;
}

void SocketServer::Client::msgavailable(void)
{
	if (!m_longpoll)
		return;
	Json::Value reply;
	if (m_firstreply) {
		reply["version"] = VERSION;
		reply["provider"] = "cfmu";
	}
	m_longpoll->unset_longpoll(get_ptr());
	reply["session"] = m_longpoll->get_session();
	m_longpoll->receive(reply, m_logfilter, m_logdiscard);
	m_longpoll.reset();
	send(reply);
}

SocketServer::Router::Router(SocketServer *server, pid_t pid, const Glib::RefPtr<Gio::Socket>& sock, const std::string& sess, int xdisplay, bool log)
	: m_server(server), m_socket(sock), m_session(sess), m_refcount(1), m_xdisplay(xdisplay), m_pid(pid), m_lifecycle(lifecycle_run), m_log(log)
{
	m_accesstime.assign_current_time();
	m_starttime = m_accesstime;
	if (true)
		std::cout << "router create " << (unsigned long long)this << " pid " << getpid() << " session " << m_session
			  << " xdisplay " << m_xdisplay << std::endl;
	if (!m_server || !m_socket || m_pid == (pid_t)-1) {
		m_server = 0;
		m_socket.reset();
		m_pid = -1;
		m_lifecycle = lifecycle_dead;
		return;
	}
	m_connin = Glib::signal_io().connect(sigc::mem_fun(*this, &Router::on_input), m_socket->get_fd(), Glib::IO_IN | Glib::IO_ERR | Glib::IO_HUP);
	m_connchildwatch = Glib::signal_child_watch().connect(sigc::mem_fun(*this, &Router::child_watch), m_pid);
	if (true)
		std::cout << "router " << (unsigned long long)this << " pid " << getpid() << " started pid " << m_pid
			  << " fd " << m_socket->get_fd() << std::endl;
}

SocketServer::Router::~Router()
{
	if (true)
		std::cout << "router destroy " << (unsigned long long)this << " session " << m_session << std::endl;
	close();
	m_connchildwatch.disconnect();
	if (m_pid != (pid_t)-1) {
		Glib::signal_child_watch().connect(sigc::ptr_fun(&Router::child_reaper), m_pid);
		m_pid = -1;
	}
}

void SocketServer::Router::reference(void) const
{
        g_atomic_int_inc(&m_refcount);
	if (false)
		std::cout << "router " << (unsigned long long)this << " pid " << getpid() << " reference " << m_refcount << std::endl;
}

void SocketServer::Router::unreference(void) const
{
        bool nz(!g_atomic_int_dec_and_test(&m_refcount));
	if (false)
		std::cout << "router " << (unsigned long long)this << " pid " << getpid() << " unreference " << m_refcount << std::endl;
	if (nz)
                return;
        delete this;
}

bool SocketServer::Router::on_output(Glib::IOCondition iocond)
{
	m_connout.disconnect();
	if (m_sendqueue.empty())
		return true;
	Json::FastWriter writer;
	std::string msg(writer.write(*m_sendqueue.begin()));
	try {
		gssize r(m_socket->send(msg.c_str(), msg.size()));
		if (r == -1) {
			if (true)
				std::cerr << "router " << (unsigned long long)this << " pid " << getpid() << " socket output return value: " << r << std::endl;
			close();
			return true;
		}
		m_sendqueue.erase(m_sendqueue.begin());
		if (m_log)
			std::cout << "router " << (unsigned long long)this << " pid " << getpid() << " tx " << msg << std::endl;
	} catch (const Gio::Error& error) {
		if (error.code () == Gio::Error::WOULD_BLOCK) {
			if (true)
				std::cerr << "router " << (unsigned long long)this << " pid " << getpid() << " socket output would block" << std::endl;
		} else {
			if (true)
				std::cerr << "router " << (unsigned long long)this << " pid " << getpid() << " socket output error: " << error.what() << std::endl;
			close();
			return true;
		}
	}
	if (!m_sendqueue.empty())
		m_connout = Glib::signal_io().connect(sigc::mem_fun(*this, &Router::on_output), m_socket->get_fd(), Glib::IO_OUT | Glib::IO_ERR);
	return true;
}

bool SocketServer::Router::on_input(Glib::IOCondition iocond)
{
	if (!m_socket) {
		m_connin.disconnect();
		return true;
	}
	if (!m_server)
		return false;
	if (m_server->is_sockclient()) {
		if (true)
			std::cerr << "router " << (unsigned long long)this << " pid " << getpid() << " receive in socket client" << std::endl;
		close();
		return false;
	}
	Json::Value root;
	try {
		char buf[262144];
 		gssize r(m_socket->receive(buf, sizeof(buf)));
		if (r <= 0) {
			if (true)
				std::cerr << "router " << (unsigned long long)this << " pid " << getpid() << " socket input return value: " << r << std::endl;
			close();
			return true;
		}
		if (r >= sizeof(buf)) {
			if (true)
				std::cerr << "router " << (unsigned long long)this << " pid " << getpid() << " socket input return value: " << r << std::endl;
			close();
			return true;
		}
		{
			Json::Reader reader;
			if (!reader.parse(buf, buf + r, root, false)) {
				if (true)
					std::cerr << "router " << (unsigned long long)this << " pid " << getpid() << " cannot parse JSON: " << std::string(buf, buf + r);
				close();
				return true;
			}
		}
		if (m_log)
			std::cout << "router " << (unsigned long long)this << " pid " << getpid() << " rx " << std::string(buf, buf + r) << std::endl;
	} catch (const Gio::Error& error) {
		if (error.code () == Gio::Error::WOULD_BLOCK) {
			if (true)
				std::cerr << "router " << (unsigned long long)this << " pid " << getpid() << " socket input would block" << std::endl;
			return true;
		} else {
			if (true)
				std::cerr << "router " << (unsigned long long)this << " pid " << getpid() << " socket input error: " << error.what() << std::endl;
			close();
			return true;
		}
	}
	if (true) {
		Glib::TimeVal tv;
		tv.assign_current_time();
		root["timestampenqueue"] = (std::string)tv.as_iso8601();
	}
	m_recvqueue.push_back(root);
	notify_longpoll();
	return true;
}

void SocketServer::Router::send(const Json::Value& v)
{
	if (!is_running())
		return;
	if (!m_socket)
		return;
	m_accesstime.assign_current_time();
	m_sendqueue.push_back(v);
	if (false) {
		Json::FastWriter writer;
		std::cout << "router " << (unsigned long long)this << " pid " << getpid() << " txqueue " << writer.write(*m_sendqueue.end()) << std::endl;
	}
	m_connout.disconnect();
	m_connout = Glib::signal_io().connect(sigc::mem_fun(*this, &Router::on_output), m_socket->get_fd(), Glib::IO_OUT | Glib::IO_ERR);
}

void SocketServer::Router::receive(Json::Value& reply, const stringset_t& logfilter, const stringset_t& logdiscard)
{
	Json::Value& replycmds(reply["cmds"]);
	unsigned int cnt(0), cntdiscarded(0), cntfiltered(0), cntpending(0);
	for (std::list<Json::Value>::iterator i(m_recvqueue.begin()), e(m_recvqueue.end()); i != e; ) {
		std::list<Json::Value>::iterator ii(i);
		++i;
		if (ii->isMember("cmdname")) {
			const Json::Value& cmdname((*ii)["cmdname"]);
			if (cmdname.isString() && cmdname.asString() == "log" && ii->isMember("item")) {
				const Json::Value& item((*ii)["item"]);
				if (item.isString()) {
					if (logdiscard.find(item.asString()) != logdiscard.end()) {
						m_recvqueue.erase(ii);
						++cntdiscarded;
						continue;
					}
					if (logfilter.find(item.asString()) != logfilter.end()) {
						++cntfiltered;
						continue;
					}
				}
			}
		}
		if (cnt >= 128) {
			++cntpending;
			continue;
		}
		replycmds.append(*ii);
		if (true) {
			Glib::TimeVal tv;
			tv.assign_current_time();
			replycmds[replycmds.size() - 1U]["timestampdequeue"] = (std::string)tv.as_iso8601();
		}
		m_recvqueue.erase(ii);
		++cnt;
	}
	if (cnt)
		m_accesstime.assign_current_time();
	reply["dequeued"] = cnt;
	reply["queuepending"] = cntpending;
	reply["queuediscarded"] = cntdiscarded;
	reply["queuefiltered"] = cntfiltered;
	if (m_lifecycle == lifecycle_status && m_recvqueue.empty()) {
		m_lifecycle = lifecycle_dead;
		if (m_server)
			m_server->reap_died_routers();
	}
}

void SocketServer::Router::close(const std::string& status)
{
	if (true) {
		std::cout << "router " << (unsigned long long)this << " pid " << getpid() << " close pid " << m_pid;
		if (m_socket)
			std::cout << " fd " << m_socket->get_fd();
		std::cout << std::endl;
	}
	m_connin.disconnect();
	m_connout.disconnect();
	m_sendqueue.clear();
	m_socket.reset();
	if (m_pid != (pid_t)-1)
		kill(m_pid, SIGHUP);
	if (m_lifecycle == lifecycle_run) {
		m_lifecycle = lifecycle_status;
		m_accesstime.assign_current_time();
		m_starttime = m_accesstime;
		Json::Value reply;
		reply["cmdname"] = "autoroute";
		reply["status"] = status.empty() ? "terminate" : status;
		m_recvqueue.push_back(reply);
	}
	notify_longpoll();
}

void SocketServer::Router::zap(void)
{
	if (true) {
		std::cout << "router " << (unsigned long long)this << " pid " << getpid() << " zap pid " << m_pid;
		if (m_socket)
			std::cout << " fd " << m_socket->get_fd();
		std::cout << std::endl;
	}
	m_connchildwatch.disconnect();
	m_connin.disconnect();
	m_connout.disconnect();
	m_sendqueue.clear();
	m_recvqueue.clear();
	if (m_pid != (pid_t)-1) {
		Glib::signal_child_watch().connect(sigc::ptr_fun(&Router::child_reaper), m_pid);
		m_pid = -1;
	}
	m_socket.reset();
	m_lifecycle = lifecycle_dead;
	notify_longpoll();
}

void SocketServer::Router::child_watch(GPid pid, int child_status)
{
	if (pid != m_pid || pid == (pid_t)-1) {
		Glib::spawn_close_pid(pid);
		return;
	}
	m_connchildwatch.disconnect();
	Glib::spawn_close_pid(pid);
	if (true)
		std::cout << "router " << (unsigned long long)this << " pid " << getpid() << " reaped pid " << m_pid
			  << " status " << child_status << std::endl;
        m_pid = -1;
	if (m_lifecycle == lifecycle_run) {
		m_lifecycle = lifecycle_status;
		m_accesstime.assign_current_time();
		m_starttime = m_accesstime;
		Json::Value reply;
		reply["cmdname"] = "autoroute";
		reply["status"] = "terminate";
		m_recvqueue.push_back(reply);
	}
	notify_longpoll();
}

void SocketServer::Router::child_reaper(GPid pid, int child_status)
{
	Glib::spawn_close_pid(pid);
}

bool SocketServer::Router::handle_timeout(const Glib::TimeVal& tvacc, const Glib::TimeVal& tvstart)
{
	if (m_lifecycle == lifecycle_dead)
		return false;
	if (tvacc < m_accesstime && tvstart < m_starttime)
		return false;
	if (m_lifecycle == lifecycle_status)
		m_lifecycle = lifecycle_dead;
	else
		close();
	return true;
}

void SocketServer::Router::set_longpoll(const Client::ptr_t& clnt)
{
	if (m_longpoll == clnt)
		return;
	notify_longpoll();
	m_longpoll = clnt;
	if (true)
		std::cout << "router " << (unsigned long long)this << " pid " << getpid() << " set longpoll client "
			  << (unsigned long)m_longpoll.operator->() << std::endl;
}

void SocketServer::Router::unset_longpoll(const Client::ptr_t& clnt)
{
	if (m_longpoll == clnt) {
		if (true)
			std::cout << "router " << (unsigned long long)this << " pid " << getpid() << " unset longpoll client "
				  << (unsigned long)m_longpoll.operator->() << std::endl;
		m_longpoll.reset();
	}
}

void SocketServer::Router::notify_longpoll(void)
{
	if (m_longpoll) {
		if (true)
			std::cout << "router " << (unsigned long long)this << " pid " << getpid() << " notify client "
				  << (unsigned long)m_longpoll.operator->() << std::endl;
		m_longpoll->msgavailable();
	}
	m_longpoll.reset();
}

SocketServer::SocketServer(CFMUAutoroute *autoroute, const Glib::RefPtr<Glib::MainLoop>& mainloop, unsigned int connlimit,
			   unsigned int actlimit, unsigned int timeout, unsigned int maxruntime, int xdisplay,
			   bool logclient, bool logrouter, bool logsockcli, bool setprocname)
	: m_autoroute(autoroute), m_mainloop(mainloop), m_connectlimit(connlimit), m_activelimit(actlimit),
	  m_timeout(timeout), m_maxruntime(maxruntime), m_xdisplay(xdisplay), m_quit(false),
	  m_logclient(logclient), m_logrouter(logrouter), m_logsockcli(logsockcli), m_setprocname(setprocname)
{
	m_cmdlist["nop"] = &SocketServer::cmd_nop;
	m_cmdlist["quit"] = &SocketServer::cmd_quit;
	m_cmdlist["departure"] = &SocketServer::cmd_departure;
	m_cmdlist["destination"] = &SocketServer::cmd_destination;
	m_cmdlist["crossing"] = &SocketServer::cmd_crossing;
	m_cmdlist["enroute"] = &SocketServer::cmd_enroute;
	m_cmdlist["levels"] = &SocketServer::cmd_levels;
	m_cmdlist["exclude"] = &SocketServer::cmd_exclude;
	m_cmdlist["tfr"] = &SocketServer::cmd_tfr;
	m_cmdlist["atmosphere"] = &SocketServer::cmd_atmosphere;
	m_cmdlist["cruise"] = &SocketServer::cmd_cruise;
	m_cmdlist["optimization"] = &SocketServer::cmd_optimization;
	m_cmdlist["preferred"] = &SocketServer::cmd_preferred;
	m_cmdlist["aircraft"] = &SocketServer::cmd_aircraft;
	m_cmdlist["start"] = &SocketServer::cmd_start;
	m_cmdlist["stop"] = &SocketServer::cmd_stop;
	m_cmdlist["continue"] = &SocketServer::cmd_continue;
	m_cmdlist["preload"] = &SocketServer::cmd_preload;
	m_cmdlist["clear"] = &SocketServer::cmd_clear;
	m_cmdlist["fplparse"] = &SocketServer::cmd_fplparse;
	m_servercmdlist["scangrib2"] = &SocketServer::servercmd_scangrib2;
	m_servercmdlist["processes"] = &SocketServer::servercmd_processes;
	m_servercmdlist["ruleinfo"] = &SocketServer::servercmd_ruleinfo;
}

SocketServer::~SocketServer()
{
	m_rtrtimeoutconn.disconnect();
	m_clients.clear();
	m_routers.clear();
	remove_listensockaddr();
}

void SocketServer::add_client(const Client::ptr_t& p)
{
	if (p)
		m_clients.insert(p);
}

void SocketServer::remove_client(const Client::ptr_t& p)
{
	clients_t::iterator i(m_clients.find(p));
	if (i != m_clients.end())
		m_clients.erase(i);
}

void SocketServer::remove_listensockaddr(void)
{
	if (!m_listensockaddr)
		return;
	if (m_listensockaddr->property_address_type() == Gio::UNIX_SOCKET_ADDRESS_PATH) {
		Glib::RefPtr<Gio::File> f(Gio::File::create_for_path(m_listensockaddr->property_path()));
		try {
			f->remove();
		} catch (...) {
		}
	}
	m_listensockaddr.reset();
}

void SocketServer::listen(const std::string& path, uid_t socketuid, gid_t socketgid, mode_t socketmode, bool sdterminate)
{
	m_listenservice.reset();
	m_listensock.reset();
	remove_listensockaddr();
	m_quit = sdterminate;
#ifdef HAVE_SYSTEMDDAEMON
	{
		int n = sd_listen_fds(0);
		if (n > 1)
			throw std::runtime_error("SystemD: too many file descriptors received");
		if (n == 1) {
			int fd = SD_LISTEN_FDS_START + 0;
			m_listensock = Gio::Socket::create_from_fd(fd);
		}
	}
#endif
	if (!m_listensock) {
		m_listensock = Gio::Socket::create(Gio::SOCKET_FAMILY_UNIX, Gio::SOCKET_TYPE_SEQPACKET, Gio::SOCKET_PROTOCOL_DEFAULT);
	        m_listensockaddr = Gio::UnixSocketAddress::create(path, Gio::UNIX_SOCKET_ADDRESS_PATH);
		if (true) {
			Glib::RefPtr<Gio::File> f(Gio::File::create_for_path(path));
			try {
				f->remove();
			} catch (...) {
			}
		}
		m_listensock->bind(m_listensockaddr, true);
		m_listensock->listen();
		//if (fchown(m_listensock->get_fd(), socketuid, socketgid))
		if (chown(path.c_str(), socketuid, socketgid))
			std::cerr << "Cannot set socket " << path << " uid/gid to " << socketuid << '/' << socketgid
				  << ": " << strerror(errno) << " (" << errno << ')' << std::endl;
		//if (fchmod(m_listensock->get_fd(), socketmode))
		if (chmod(path.c_str(), socketmode))
			std::cerr << "Cannot set socket " << path << " mode to " << std::oct << '0'
				  << (unsigned int)socketmode << std::dec << ": " << strerror(errno) << " (" << errno << ')' << std::endl;
		m_quit = false;
	}
	m_listenservice = Gio::SocketService::create();
	m_listenservice->signal_incoming().connect(sigc::mem_fun(*this, &SocketServer::on_connect));
	if (!m_listenservice->add_socket(m_listensock))
		throw std::runtime_error("Cannot listen to socket");
	m_listenservice->start();
}

bool SocketServer::on_connect(const Glib::RefPtr<Gio::SocketConnection>& connection, const Glib::RefPtr<Glib::Object>& source_object)
{
	Client::ptr_t p(new Client(this, connection, m_logclient));
	return true;
}

void SocketServer::count_routers(unsigned int& total, unsigned int& active)
{
	unsigned int t(0), a(0);
	for (routers_t::iterator i(m_routers.begin()), e(m_routers.end()); i != e;) {
		if (!i->second || i->second->is_dead()) {
			if (true)
				std::cout << "router " << (unsigned long)i->second.operator->() << " session " << i->first << " remove" << std::endl;
			routers_t::iterator ii(i);
			++i;
			m_routers.erase(ii);
			continue;
		}
		++t;
		if (i->second->is_running())
			++a;
		++i;
	}
	if (m_routers.empty())
		m_rtrtimeoutconn.disconnect();
	total = t;
	active = a;
}

SocketServer::Router::ptr_t SocketServer::find_router(const std::string& session, bool create)
{
	if (session.empty())
		return Router::ptr_t();
	if (true) {
		std::cout << "find_router: pid " << getpid() << std::endl;
		for (routers_t::iterator i(m_routers.begin()), e(m_routers.end()); i != e; ++i)
			std::cout << "find_router " << (unsigned long)i->second.operator->() << " session " << i->first << std::endl;
	}
	routers_t::iterator i(m_routers.find(session));
	if (i != m_routers.end()) {
		if (i->second && !i->second->is_dead())
			return i->second;
		if (true)
			std::cout << "router " << (unsigned long)i->second.operator->() << " session " << i->first << " remove" << std::endl;
		m_routers.erase(i);
		if (m_routers.empty())
			m_rtrtimeoutconn.disconnect();
	}
	if (!create)
		return Router::ptr_t();
	{
		unsigned int total, active;
		count_routers(total, active);
		if (!create || total >= m_connectlimit || active >= m_activelimit)
			return Router::ptr_t();
	}
	// find xdisplay
	int xdisplay(-1);
	if (m_xdisplay >= 0) {
		typedef std::set<int> xdisps_t;
		xdisps_t xdisps;
		for (int i(m_xdisplay), e(m_xdisplay + m_routers.size()); i <= e; ++i)
			xdisps.insert(i);
		for (routers_t::const_iterator i(m_routers.begin()), e(m_routers.end()); i != e; ++i)
			xdisps.erase(i->second->get_xdisplay());
		if (!xdisps.empty())
			xdisplay = *xdisps.begin();
	}
	// create socketpair
	int socks[2];
	if (socketpair(AF_LOCAL, SOCK_SEQPACKET, 0, socks))
		return Router::ptr_t();
	// fork
	pid_t pid = fork();
	if (!pid) {
		// child
		close(socks[0]);
		m_listenservice->stop();
		m_listensockaddr.reset();
		m_listenservice.reset();
		m_listensock.reset();
		if (true)
			std::cout << "New session " << session << " kill clients" << std::endl;
		for (clients_t::iterator i(m_clients.begin()), e(m_clients.end()); i != e; ) {
			clients_t::iterator i2(i);
			++i;
			(*i2)->close();
		}
		if (true)
			std::cout << "New session " << session << " kill routers" << std::endl;
		while (!m_routers.empty()) {
			m_routers.begin()->second->zap();
			m_routers.erase(m_routers.begin());
		}
		if (true)
			std::cout << "New session " << session << " start sockclient" << std::endl;
		if (m_setprocname) {
#ifdef HAVE_SYS_PRCTL_H
			prctl(PR_SET_NAME, (unsigned long)session.c_str(), 0, 0, 0);
#endif
		}
		sockclient(socks[1], xdisplay);
		return Router::ptr_t();
	}
	close(socks[1]);
	if (pid == -1) {
		close(socks[0]);
		return Router::ptr_t();
	}
	if (m_routers.empty()) {
		m_rtrtimeoutconn.disconnect();
		m_rtrtimeoutconn = Glib::signal_timeout().connect_seconds(sigc::mem_fun(*this, &SocketServer::router_timeout_handler), std::min(std::max(m_timeout / 10U, 1U), 10U));
	}
	Router::ptr_t p(new Router(this, pid, Gio::Socket::create_from_fd(socks[0]), session, xdisplay, m_logrouter));
	m_routers[session] = p;
	if (true) {
		std::cout << "find_router: after creation: pid " << getpid() << std::endl;
		for (routers_t::iterator i(m_routers.begin()), e(m_routers.end()); i != e; ++i)
			std::cout << "find_router " << (unsigned long)i->second.operator->() << " session " << i->first << std::endl;
	}
	return p;
}

bool SocketServer::stop_router(const std::string& session)
{
	Router::ptr_t p(find_router(session, false));
	if (!p)
		return false;
	p->close();
	return true;
}

void SocketServer::reap_died_routers(void)
{
	for (routers_t::iterator i(m_routers.begin()), e(m_routers.end()); i != e;) {
		routers_t::iterator ii(i);
		++i;
		if (ii->second && !ii->second->is_dead())
			continue;
		if (true)
			std::cout << "router " << (unsigned long)ii->second.operator->() << " session " << ii->first << " remove" << std::endl;
		m_routers.erase(ii);
	}
	if (m_routers.empty()) {
		m_rtrtimeoutconn.disconnect();
		if (m_quit)
			m_mainloop->quit();
	}
}

bool SocketServer::router_timeout_handler(void)
{
	Glib::TimeVal tvacc, tvrun;
	tvacc.assign_current_time();
	tvrun = tvacc;
	tvacc.subtract_seconds(m_timeout);
	tvrun.subtract_seconds(m_maxruntime);
	for (routers_t::iterator i(m_routers.begin()), e(m_routers.end()); i != e;) {
		Router::ptr_t p(i->second);
		++i;
		p->handle_timeout(tvacc, tvrun);
	}
	return true;
}

void SocketServer::sockclient(int fd, int xdisplay)
{
	if (!m_autoroute)
		return;
	m_autoroute->set_validator_xdisplay(xdisplay);
	m_listensock = Gio::Socket::create_from_fd(fd);
        m_listensock->property_blocking() = true;
	Glib::signal_io().connect(sigc::mem_fun(*this, &SocketServer::sockcli_input), m_listensock->get_fd(), Glib::IO_IN | Glib::IO_ERR | Glib::IO_HUP);
	m_autoroute->signal_statuschange().clear();
	m_autoroute->signal_log().clear();
	m_autoroute->signal_statuschange().connect(sigc::mem_fun(*this, &SocketServer::sockcli_statuschange));
	m_autoroute->signal_log().connect(sigc::mem_fun(*this, &SocketServer::sockcli_autoroutelog));
	if (m_xdisplay >= 0) {
		std::ostringstream disp;
		disp << ':' << m_xdisplay;
		Glib::setenv("DISPLAY", disp.str(), true);
	}
}

void SocketServer::sockcli_close(void)
{
	m_listensock.reset();
	if (m_mainloop)
		m_mainloop->quit();
}

void SocketServer::sockcli_send(const Json::Value& v)
{
	Json::FastWriter writer;
	std::string msg(writer.write(v));
	try {
		gssize r(m_listensock->send(msg.c_str(), msg.size()));
		if (r == -1) {
			if (true)
				std::cerr << "sockcli " << getpid() << " socket output return value: " << r << std::endl;
			sockcli_close();
			return;
		}
		if (m_logsockcli)
			std::cerr << "sockcli " << getpid() << " tx " << msg << std::endl;
	} catch (const Gio::Error& error) {
		if (true)
			std::cerr << "sockcli " << getpid() << " socket output error: " << error.what() << std::endl;
		sockcli_close();
		return;
	}
}

void SocketServer::sockcli_timestamp(Json::Value& v)
{
	Glib::TimeVal tv;
	tv.assign_current_time();
	v["timestamp"] = (std::string)tv.as_iso8601();
}

bool SocketServer::sockcli_input(Glib::IOCondition iocond)
{
	Json::Value root;
	Json::Value reply;
	try {
		char buf[262144];
 		gssize r(m_listensock->receive(buf, sizeof(buf)));
		if (r <= 0) {
			if (true)
				std::cerr << "sockcli " << getpid() << " socket input return value: " << r << std::endl;
			sockcli_close();
			return true;
		}
		if (r >= sizeof(buf)) {
			if (true)
				std::cerr << "sockcli " << getpid() << " socket input return value: " << r << std::endl;
			reply["error"] = std::string("invalid message size");
			sockcli_timestamp(reply);
			sockcli_send(reply);
			return true;
		}
		{
			Json::Reader reader;
			if (!reader.parse(buf, buf + r, root, false)) {
				if (true)
					std::cerr << "sockcli " << getpid() << " cannot parse JSON: " << std::string(buf, buf + r);
				reply["error"] = std::string("cannot parse JSON");
				sockcli_timestamp(reply);
				sockcli_send(reply);
				return true;
			}
		}
		if (m_logsockcli)
			std::cout << "sockcli " << getpid() << " rx " << std::string(buf, buf + r) << std::endl;
	} catch (const Gio::Error& error) {
		if (error.code () == Gio::Error::WOULD_BLOCK) {
			if (true)
				std::cerr << "sockcli " << getpid() << " socket input would block" << std::endl;
			return true;
		} else {
			if (true)
				std::cerr << "sockcli " << getpid() << " socket input error: " << error.what() << std::endl;
			sockcli_close();
			return true;
		}
	}
	try {
		if (root.isMember("cmdseq"))
			reply["cmdseq"] = root["cmdseq"];
		std::string cmdname(root.get("cmdname", "").asString());
		reply["cmdname"] = cmdname.empty() ? "autoroute" : cmdname;
		Json::Value reply1(reply);
		cmdlist_t::const_iterator cmdi(m_cmdlist.find(cmdname));
                if (cmdi == m_cmdlist.end()) {
			reply1["error"] = "command not found";
                } else {
                        (this->*(cmdi->second))(root, reply1);
                }
		sockcli_timestamp(reply1);
                sockcli_send(reply1);
        } catch (const std::runtime_error& e) {
		reply["error"] = e.what();
		sockcli_timestamp(reply);
               sockcli_send(reply);
        } catch (const Glib::Exception& e) {
		reply["error"] = (std::string)e.what();
		sockcli_timestamp(reply);
                sockcli_send(reply);
        }
	return true;
}

void SocketServer::sockcli_statuschange(CFMUAutoroute::statusmask_t status)
{
        if (status & (CFMUAutoroute::statusmask_starting |
                      CFMUAutoroute::statusmask_stoppingdone |
                      CFMUAutoroute::statusmask_stoppingerror)) {
		Json::Value reply;
		reply["cmdname"] = "autoroute";
                if (status & CFMUAutoroute::statusmask_starting) {
                        reply["status"] =  "starting";
                } else if (status & (CFMUAutoroute::statusmask_stoppingdone |
                                     CFMUAutoroute::statusmask_stoppingerror)) {
                        reply["status"] = "stopping";
                        reply["routesuccess"] = !!(status & CFMUAutoroute::statusmask_stoppingdone);
			reply["localiteration"] = m_autoroute->get_local_iterationcount();
			reply["remoteiteration"] = m_autoroute->get_remote_iterationcount();
			reply["wallclocktime"] = m_autoroute->get_wallclocktime().as_double();
			reply["validatortime"] = m_autoroute->get_validatortime().as_double();
                }
		sockcli_timestamp(reply);
                sockcli_send(reply);
        }
}

void SocketServer::sockcli_autoroutelog(CFMUAutoroute::log_t item, const std::string& line)
{
	Json::Value reply;
        if (item != CFMUAutoroute::log_fplproposal) {
	        reply["cmdname"] = "log";
                reply["item"] = (std::string)to_str(item);
                reply["text"] = (std::string)line;
		sockcli_timestamp(reply);
                sockcli_send(reply);
                return;
        }
        // new flight plan
	reply["cmdname"] = "fpl";
	if (m_autoroute) {
		const FPlanRoute& route(m_autoroute->get_route());
		encode_fpl(reply, route);
		reply["mintime"] = m_autoroute->get_mintime(true);
		reply["routetime"] = m_autoroute->get_routetime(true);
		reply["minfuel"] = m_autoroute->get_minfuel(true);
		reply["routefuel"] = m_autoroute->get_routefuel(true);
		reply["mintimezerowind"] = m_autoroute->get_mintime(false);
		reply["routetimezerowind"] = m_autoroute->get_routetime(false);
		reply["minfuelzerowind"] = m_autoroute->get_minfuel(false);
		reply["routefuelzerowind"] = m_autoroute->get_routefuel(false);
		reply["fpl"] = (std::string)m_autoroute->get_simplefpl();
		reply["iteration"] = m_autoroute->get_iterationcount();
		reply["localiteration"] = m_autoroute->get_local_iterationcount();
		reply["remoteiteration"] = m_autoroute->get_remote_iterationcount();
		reply["wallclocktime"] = m_autoroute->get_wallclocktime().as_double();
		reply["validatortime"] = m_autoroute->get_validatortime().as_double();
	} else {
		reply["error"] = "no autoroute instance";
	}
	sockcli_timestamp(reply);
	sockcli_send(reply);
}

Point SocketServer::get_point(const Json::Value& cmd, const std::string& name)
{
	Point pt;
	pt.set_invalid();
	std::string s;
	if (cmd.isMember(name))
		s = cmd.get(name, "").asString();
	std::string lat, lon;
	if (s.empty()) {
		{
			const Json::Value& lat(cmd[name + "lat"]);
			const Json::Value& lon(cmd[name + "lon"]);
			if (lat.isIntegral() && lon.isIntegral()) {
				pt.set_lat(lat.asInt());
				pt.set_lon(lon.asInt());
				return pt;
			}
		}
		{
			const Json::Value& lat(cmd[name + "latdeg"]);
			const Json::Value& lon(cmd[name + "londeg"]);
			if (lat.isNumeric() && lon.isNumeric()) {
				pt.set_lat_deg_dbl(lat.asDouble());
				pt.set_lon_deg_dbl(lon.asDouble());
				return pt;
			}
		}
		lat = cmd.get(name + "latstr", "").asString();
		lon = cmd.get(name + "lonstr", "").asString();
		unsigned int x(pt.set_str(lat) | pt.set_str(lon));
		if (!(x & ~(Point::setstr_lat | Point::setstr_lon)))
			return pt;
		pt.set_invalid();
		return pt;
	} else {
		std::string::size_type n(s.find(','));
		if (n == std::string::npos)
			return pt;
		lat = s.substr(0, n);
		lon = s.substr(n + 1);
	}
	if (lat.empty() || lon.empty())
		return pt;
	char *cp1(0), *cp2(0);
	pt.set_lat(strtol(lat.c_str(), &cp1, 10));
	pt.set_lon(strtol(lon.c_str(), &cp2, 10));
	if (*cp1 || *cp2)
		pt.set_invalid();
	return pt;
}

void SocketServer::set_point(Json::Value& cmd, const std::string& name, const Point& pt)
{
	std::ostringstream oss;
	if (!pt.is_invalid()) {
		oss << pt.get_lat() << ',' << pt.get_lon();
		cmd[name + "lat"] = pt.get_lat();
		cmd[name + "lon"] = pt.get_lon();
		cmd[name + "latdeg"] = pt.get_lat_deg_dbl();
		cmd[name + "londeg"] = pt.get_lon_deg_dbl();
		cmd[name + "latstr"] = (std::string)pt.get_lat_str2();
		cmd[name + "lonstr"] = (std::string)pt.get_lon_str2();
	}
	cmd[name] = oss.str();
}

void SocketServer::encode_wpt(Json::Value& fplwpt, const FPlanWaypoint& wpt)
{
	fplwpt["icao"] = (std::string)wpt.get_icao();
	fplwpt["name"] = (std::string)wpt.get_name();
	fplwpt["pathname"] = (std::string)wpt.get_pathname();
	fplwpt["pathcode"] = (std::string)wpt.get_pathcode_name();
	fplwpt["note"] = (std::string)wpt.get_note();
	fplwpt["time"] = (Json::LargestInt)wpt.get_time_unix();
	fplwpt["save_time"] = (Json::LargestInt)wpt.get_save_time_unix();
	fplwpt["flighttime"] = (Json::LargestInt)wpt.get_flighttime();
	fplwpt["frequency"] = wpt.get_frequency();
	set_point(fplwpt, "coord", wpt.get_coord());
	fplwpt["altitude"] = wpt.get_altitude();
	fplwpt["flags"] = wpt.get_flags();
	fplwpt["flightrules"] = wpt.is_ifr() ? "ifr" : "vfr";
	fplwpt["altmode"] = wpt.is_standard() ? "std" : "qnh";
	fplwpt["climb"] = wpt.is_climb();
	fplwpt["descent"] = wpt.is_descent();
	fplwpt["turnpoint"] = wpt.is_turnpoint();
	fplwpt["winddir"] = wpt.get_winddir_deg();
	fplwpt["windspeed"] = wpt.get_windspeed_kts();
	fplwpt["qff"] = wpt.get_qff_hpa();
	fplwpt["isaoffset"] = wpt.get_isaoffset_kelvin();
	fplwpt["truealt"] = wpt.get_truealt();
	fplwpt["truetrack"] = wpt.get_truetrack_deg();
	fplwpt["trueheading"] = wpt.get_trueheading_deg();
	fplwpt["declination"] = wpt.get_declination_deg();
	fplwpt["magnetictrack"] = wpt.get_magnetictrack_deg();
	fplwpt["magneticheading"] = wpt.get_magneticheading_deg();
	fplwpt["dist"] = wpt.get_dist_nmi();
	fplwpt["fuel"] = wpt.get_fuel_usg();
	fplwpt["tas"] = wpt.get_tas_kts();
	fplwpt["rpm"] = wpt.get_rpm();
	fplwpt["mp"] = wpt.get_mp_inhg();
	fplwpt["type"] = wpt.get_type_string();
	{
		Wind<double> wind;
		wind.set_wind(wpt.get_winddir_deg(), wpt.get_windspeed_kts());
		wind.set_crs_tas(wpt.get_truetrack_deg(), wpt.get_tas_kts());
		if (!std::isnan(wind.get_gs()))
			fplwpt["gs"] = wind.get_gs();
	}
}

void SocketServer::encode_fpl(Json::Value& reply, const FPlanRoute& route)
{
	reply["note"] = (std::string)route.get_note();
	reply["curwpt"] = route.get_curwpt();
	reply["offblock"] = (Json::LargestInt)route.get_time_offblock_unix();
	reply["save_offblock"] = (Json::LargestInt)route.get_save_time_offblock_unix();
	reply["onblock"] = (Json::LargestInt)route.get_time_onblock_unix();
	reply["save_onblock"] = (Json::LargestInt)route.get_save_time_onblock_unix();
	{
		Json::Value& fplan(reply["fplan"]);
		for (unsigned int i = 0; i < route.get_nrwpt(); ++i) {
			const FPlanWaypoint& wpt(route[i]);
			Json::Value fplwpt;
			encode_wpt(fplwpt, wpt);
			if (m_autoroute) {
				int alt(wpt.get_altitude());
				double densityalt, tas;
				m_autoroute->get_cruise_performance(alt, densityalt, tas);
				if (alt == wpt.get_altitude()) {
					if (!std::isnan(densityalt))
						fplwpt["densityalt"] = densityalt;
				}
			}
			fplan.append(fplwpt);
		}
	}
	reply["gcdist"] = route.gc_distance_nmi_dbl();
	reply["routedist"] = route.total_distance_nmi_dbl();
}

void SocketServer::decode_fpl(FPlanRoute& route, const Json::Value& json)
{
	if (json.isMember("note") && json["note"].isString())
		route.set_note(json["note"].asString());
	if (json.isMember("curwpt") && json["curwpt"].isIntegral())
		route.set_curwpt(json["curwpt"].asInt());
	if (json.isMember("offblock") && json["offblock"].isIntegral())
		route.set_time_offblock_unix(json["offblock"].asInt());
	if (json.isMember("save_offblock") && json["save_offblock"].isIntegral())
		route.set_save_time_offblock_unix(json["save_offblock"].asInt());
	if (json.isMember("onblock") && json["onblock"].isIntegral())
		route.set_time_onblock_unix(json["onblock"].asInt());
	if (json.isMember("save_onblock") && json["save_onblock"].isIntegral())
		route.set_save_time_onblock_unix(json["save_onblock"].asInt());
	if (json.isMember("fplan") && json["fplan"].isArray()) {
		const Json::Value& fplan(json["fplan"]);
		for (unsigned int i = 0; i < fplan.size(); ++i) {
			const Json::Value& fplwpt(fplan[i]);
			 FPlanWaypoint wpt;
			 if (fplwpt.isMember("icao") && fplwpt["icao"].isString())
				 wpt.set_icao(fplwpt["icao"].asString());
			 if (fplwpt.isMember("name") && fplwpt["name"].isString())
				 wpt.set_name(fplwpt["name"].asString());
			 if (fplwpt.isMember("pathname") && fplwpt["pathname"].isString())
				 wpt.set_pathname(fplwpt["pathname"].asString());
			 if (fplwpt.isMember("pathcode") && fplwpt["pathcode"].isString())
				 wpt.set_pathcode_name(fplwpt["pathcode"].asString());
			 if (fplwpt.isMember("note") && fplwpt["note"].isString())
				 wpt.set_note(fplwpt["note"].asString());
			 if (fplwpt.isMember("time") && fplwpt["time"].isIntegral())
				 wpt.set_time_unix(fplwpt["time"].asInt());
			 if (fplwpt.isMember("save_time") && fplwpt["save_time"].isIntegral())
				 wpt.set_save_time_unix(fplwpt["save_time"].asInt());
			 if (fplwpt.isMember("flighttime") && fplwpt["flighttime"].isIntegral())
				 wpt.set_flighttime(fplwpt["flighttime"].asInt());
			 if (fplwpt.isMember("frequency") && fplwpt["frequency"].isIntegral())
				 wpt.set_frequency(fplwpt["frequency"].asInt());
			 wpt.set_coord(get_point(fplwpt, "coord"));
			 if (fplwpt.isMember("altitude") && fplwpt["altitude"].isIntegral())
				 wpt.set_altitude(fplwpt["altitude"].asInt());
			 if (fplwpt.isMember("flags") && fplwpt["flags"].isIntegral())
				 wpt.set_flags(fplwpt["flags"].asInt());
			 if (fplwpt.isMember("flightrules") && fplwpt["flightrules"].isString() &&
			     (fplwpt["flightrules"].asString() == "vfr" || fplwpt["flightrules"].asString() == "ifr"))
				 wpt.frob_flags((fplwpt["flightrules"].asString() == "ifr") ? FPlanWaypoint::altflag_ifr : 0, FPlanWaypoint::altflag_ifr);
			 if (fplwpt.isMember("altmode") && fplwpt["altmode"].isString() &&
			     (fplwpt["altmode"].asString() == "qnh" || fplwpt["altmode"].asString() == "std"))
				 wpt.frob_flags((fplwpt["altmode"].asString() == "std") ? FPlanWaypoint::altflag_standard : 0, FPlanWaypoint::altflag_standard);
			 if (fplwpt.isMember("climb") && fplwpt["climb"].isBool())
				 wpt.frob_flags(fplwpt["climb"].asBool() ? FPlanWaypoint::altflag_climb : 0, FPlanWaypoint::altflag_climb);
			 if (fplwpt.isMember("descent") && fplwpt["descent"].isBool())
				 wpt.frob_flags(fplwpt["descent"].asBool() ? FPlanWaypoint::altflag_descent : 0, FPlanWaypoint::altflag_descent);
			 if (fplwpt.isMember("turnpoint") && fplwpt["turnpoint"].isBool())
				 wpt.frob_flags(fplwpt["turnpoint"].asBool() ? FPlanWaypoint::altflag_turnpoint : 0, FPlanWaypoint::altflag_turnpoint);
			 if (fplwpt.isMember("winddir") && fplwpt["winddir"].isNumeric())
				 wpt.set_winddir_deg(fplwpt["winddir"].asDouble());
			 if (fplwpt.isMember("windspeed") && fplwpt["windspeed"].isNumeric())
				 wpt.set_windspeed_kts(fplwpt["windspeed"].asDouble());
			 if (fplwpt.isMember("qff") && fplwpt["qff"].isNumeric())
				 wpt.set_qff_hpa(fplwpt["qff"].asDouble());
			 if (fplwpt.isMember("isaoffset") && fplwpt["isaoffset"].isNumeric())
				 wpt.set_isaoffset_kelvin(fplwpt["isaoffset"].asDouble());
			 if (fplwpt.isMember("truealt") && fplwpt["truealt"].isNumeric())
				 wpt.set_truealt(fplwpt["truealt"].asDouble());
			 if (fplwpt.isMember("trueheading") && fplwpt["trueheading"].isNumeric())
				 wpt.set_trueheading_deg(fplwpt["trueheading"].asDouble());
			 if (fplwpt.isMember("declination") && fplwpt["declination"].isNumeric())
				 wpt.set_declination_deg(fplwpt["declination"].asDouble());
			 if (fplwpt.isMember("fuel") && fplwpt["fuel"].isNumeric())
				 wpt.set_fuel_usg(fplwpt["fuel"].asDouble());
			 if (fplwpt.isMember("tas") && fplwpt["tas"].isNumeric())
				 wpt.set_tas_kts(fplwpt["tas"].asDouble());
			 if (fplwpt.isMember("rpm") && fplwpt["rpm"].isNumeric())
				 wpt.set_rpm(fplwpt["rpm"].asDouble());
			 if (fplwpt.isMember("mp") && fplwpt["mp"].isNumeric())
				 wpt.set_mp_inhg(fplwpt["mp"].asDouble());
			 if (fplwpt.isMember("type") && fplwpt["type"].isString())
				 wpt.set_type_string(fplwpt["type"].asString());
			 route.insert_wpt(~0U, wpt);
		}
	}
	route.recompute_dist();
}

void SocketServer::encode_rule(Json::Value& reply, const TrafficFlowRestrictions::RuleResult& rule)
{
	switch (rule.get_codetype()) {
	case TrafficFlowRestrictions::RuleResult::codetype_mandatory:
		reply["codetype"] = "mandatory";
		break;

	case TrafficFlowRestrictions::RuleResult::codetype_forbidden:
		reply["codetype"] = "forbidden";
		break;

	case TrafficFlowRestrictions::RuleResult::codetype_closed:
		reply["codetype"] = "closedforcruising";
		break;

	default:
		reply["codetype"] = "?";
		break;
	}
	reply["dct"] = rule.is_dct();
	reply["unconditional"] = rule.is_unconditional();
	reply["routestatic"] = rule.is_routestatic();
	reply["mandatoryinbound"] = rule.is_mandatoryinbound();
	reply["disabled"] = rule.is_disabled();
	{
		Json::Value& vset(reply["vertices"]);
		for (TrafficFlowRestrictions::RuleResult::set_t::const_iterator i(rule.get_vertexset().begin()), e(rule.get_vertexset().end()); i != e; ++i)
			vset.append(Json::Value(*i));
		Json::Value& eset(reply["edges"]);
		for (TrafficFlowRestrictions::RuleResult::set_t::const_iterator i(rule.get_edgeset().begin()), e(rule.get_edgeset().end()); i != e; ++i)
			eset.append(Json::Value(*i));
		Json::Value& rset(reply["reflocs"]);
		for (TrafficFlowRestrictions::RuleResult::set_t::const_iterator i(rule.get_reflocset().begin()), e(rule.get_reflocset().end()); i != e; ++i)
			rset.append(Json::Value(*i));
	}
	reply["desc"] = rule.get_desc();
	reply["oprgoal"] = rule.get_oprgoal();
	reply["cond"] = rule.get_cond();
	Json::Value& ralternatives(reply["alternatives"]);
	for (unsigned int i = 0; i < rule.size(); ++i) {
		const TrafficFlowRestrictions::RuleResult::Alternative& alt(rule[i]);
		Json::Value ralt;
		{
			Json::Value& vset(ralt["vertices"]);
			for (TrafficFlowRestrictions::RuleResult::set_t::const_iterator i(alt.get_vertexset().begin()), e(alt.get_vertexset().end()); i != e; ++i)
				vset.append(Json::Value(*i));
			Json::Value& eset(ralt["edges"]);
			for (TrafficFlowRestrictions::RuleResult::set_t::const_iterator i(alt.get_edgeset().begin()), e(alt.get_edgeset().end()); i != e; ++i)
				eset.append(Json::Value(*i));
		}
		Json::Value& rsequences(reply["sequences"]);
		for (unsigned int j = 0; j < alt.size(); ++j) {
			const TrafficFlowRestrictions::RuleResult::Alternative::Sequence& el(alt[j]);
			Json::Value rseq;
			rseq["type"] = el.get_type_string();
			switch (el.get_type()) {
			case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_sid:
			case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_star:
				rseq["ident"] = el.get_ident();
				rseq["airport"] = el.get_airport();
				break;

			case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_airspace:
				rseq["ident"] = el.get_ident();
				rseq["class"] = (std::string)AirspacesDb::Airspace::get_class_string(el.get_bdryclass(), AirspacesDb::Airspace::typecode_ead);
				break;

			case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_airway:
				rseq["ident"] = el.get_ident();
				// fall through

			case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_dct:
			{
				const TrafficFlowRestrictions::RuleResult::Alternative::Sequence::RulePoint& pt(el.get_point1());
				Json::Value& rpt(rseq["point1"]);
				rpt["ident"] = pt.get_ident();
				set_point(rpt, "coord", pt.get_coord());
				rpt["type"] = to_str(pt.get_type());
				rpt["vor"] = pt.is_vor();
				rpt["ndb"] = pt.is_ndb();
				rpt["valid"] = pt.is_valid();
				// fall through
			}

			case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_point:
			{
				const TrafficFlowRestrictions::RuleResult::Alternative::Sequence::RulePoint& pt(el.get_point0());
				Json::Value& rpt(rseq["point0"]);
				rpt["ident"] = pt.get_ident();
				set_point(rpt, "coord", pt.get_coord());
				rpt["type"] = to_str(pt.get_type());
				rpt["vor"] = pt.is_vor();
				rpt["ndb"] = pt.is_ndb();
				rpt["valid"] = pt.is_valid();
				break;
			}

			default:
				break;
			}
			if (el.is_lower_alt_valid())
				rseq["altlower"] = el.get_lower_alt();
			if (el.is_upper_alt_valid())
				rseq["altupper"] = el.get_upper_alt();
			rsequences.append(rseq);
		}
		ralternatives.append(ralt);
	}
}

void SocketServer::cmd_nop(const Json::Value& cmdin, Json::Value& cmdout)
{
}

void SocketServer::cmd_quit(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (m_mainloop)
		m_mainloop->quit();
}

void SocketServer::cmd_departure(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (!m_autoroute) {
		cmdout["error"] = "no autoroute instance";
		return;
	}
	if (cmdin.isMember("icao") || cmdin.isMember("name")) {
		std::string icao(cmdin.get("icao", "").asString());
		std::string name(cmdin.get("name", "").asString());
		if (!m_autoroute->set_departure(icao, name))
			throw std::runtime_error("departure aerodrome " + icao + " " + name + " not found");
	}
	if (cmdin.isMember("ifr")) {
		const Json::Value& x(cmdin["ifr"]);
		m_autoroute->set_departure_ifr(!x.isBool() || x.asBool());
	}
	if (cmdin.isMember("vfr")) {
		const Json::Value& x(cmdin["vfr"]);
		m_autoroute->set_departure_ifr(x.isBool() && !x.asBool());
	}
	{
		Point pt(get_point(cmdin, "sid"));
		if (pt.is_invalid()) {
			if (cmdin.isMember("sidident"))
				m_autoroute->set_sid(cmdin.get("sidident", "").asString());
		} else {
			m_autoroute->set_sid(pt);
		}
	}
	if (cmdin.isMember("sidlimit"))
		m_autoroute->set_sidlimit(cmdin["sidlimit"].asDouble());
	if (cmdin.isMember("sidpenalty"))
		m_autoroute->set_sidpenalty(cmdin["sidpenalty"].asDouble());
	if (cmdin.isMember("sidoffset"))
		m_autoroute->set_sidoffset(cmdin["sidoffset"].asDouble());
	if (cmdin.isMember("sidminimum"))
		m_autoroute->set_sidminimum(cmdin["sidminimum"].asDouble());
	if (cmdin.isMember("siddb"))
		m_autoroute->set_siddb(cmdin["siddb"].asBool());
	if (cmdin.isMember("time"))
		m_autoroute->set_deptime(cmdin["time"].asUInt());
	if (cmdin.isMember("date")) {
		Glib::TimeVal tv;
		if (tv.assign_from_iso8601(cmdin["date"].asString()))
			m_autoroute->set_deptime(tv.tv_sec);
	}
	if (m_autoroute->get_departure().is_valid()) {
		cmdout["icao"] = (std::string)m_autoroute->get_departure().get_icao();
		cmdout["name"] = (std::string)m_autoroute->get_departure().get_name();
		set_point(cmdout, "coord", m_autoroute->get_departure().get_coord());
	}
	if (m_autoroute->get_sid().is_invalid()) {
		cmdout.removeMember("sid");
	} else {
	        set_point(cmdout, "sid", m_autoroute->get_sid());
		cmdout["sidtype"] = to_str(m_autoroute->get_sidtype());
		cmdout["sidident"] = m_autoroute->get_sidident();
	}
	cmdout["sidlimit"] = m_autoroute->get_sidlimit();
	cmdout["sidpenalty"] = m_autoroute->get_sidpenalty();
	cmdout["sidoffset"] = m_autoroute->get_sidoffset();
	cmdout["sidminimum"] = m_autoroute->get_sidminimum();
	cmdout["siddb"] = m_autoroute->get_siddb();
	if (m_autoroute->get_departure_ifr()) {		
		cmdout["ifr"] = true;
		cmdout.removeMember("vfr");
	} else {		
		cmdout["vfr"] = true;
		cmdout.removeMember("ifr");
	}
	cmdout["time"] = (Json::LargestInt)m_autoroute->get_deptime();
	{
		Glib::TimeVal tv(m_autoroute->get_deptime(), 0);
		cmdout["date"] = (std::string)tv.as_iso8601();
	}
}

void SocketServer::cmd_destination(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (!m_autoroute) {
		cmdout["error"] = "no autoroute instance";
		return;
	}
	if (cmdin.isMember("icao") || cmdin.isMember("name")) {
		std::string icao(cmdin.get("icao", "").asString());
		std::string name(cmdin.get("name", "").asString());
		if (!m_autoroute->set_destination(icao, name))
			throw std::runtime_error("destination aerodrome " + icao + " " + name + " not found");
	}
	if (cmdin.isMember("ifr")) {
		const Json::Value& x(cmdin["ifr"]);
		m_autoroute->set_destination_ifr(!x.isBool() || x.asBool());
	}
	if (cmdin.isMember("vfr")) {
		const Json::Value& x(cmdin["vfr"]);
		m_autoroute->set_destination_ifr(x.isBool() && !x.asBool());
	}
	{
		Point pt(get_point(cmdin, "star"));
		if (pt.is_invalid()) {
			if (cmdin.isMember("starident"))
				m_autoroute->set_star(cmdin.get("starident", "").asString());
		} else {
			m_autoroute->set_star(pt);
		}
	}
	if (cmdin.isMember("starlimit"))
		m_autoroute->set_starlimit(cmdin["starlimit"].asDouble());
	if (cmdin.isMember("starpenalty"))
		m_autoroute->set_starpenalty(cmdin["starpenalty"].asDouble());
	if (cmdin.isMember("staroffset"))
		m_autoroute->set_staroffset(cmdin["staroffset"].asDouble());
	if (cmdin.isMember("starminimum"))
		m_autoroute->set_starminimum(cmdin["starminimum"].asDouble());
	if (cmdin.isMember("stardb"))
		m_autoroute->set_stardb(cmdin["stardb"].asBool());
	if (m_autoroute->get_destination().is_valid()) {
		cmdout["icao"] = (std::string)m_autoroute->get_destination().get_icao();
		cmdout["name"] = (std::string)m_autoroute->get_destination().get_name();
		set_point(cmdout, "coord", m_autoroute->get_destination().get_coord());
	}
	if (m_autoroute->get_star().is_invalid()) {
		cmdout.removeMember("star");
	} else {
	        set_point(cmdout, "star", m_autoroute->get_star());
		cmdout["startype"] = to_str(m_autoroute->get_startype());
		cmdout["starident"] = m_autoroute->get_starident();
	}
	cmdout["starlimit"] = m_autoroute->get_starlimit();
	cmdout["starpenalty"] = m_autoroute->get_starpenalty();
	cmdout["staroffset"] = m_autoroute->get_staroffset();
	cmdout["starminimum"] = m_autoroute->get_starminimum();
	cmdout["stardb"] = m_autoroute->get_stardb();
	if (m_autoroute->get_destination_ifr()) {		
		cmdout["ifr"] = true;
		cmdout.removeMember("vfr");
	} else {		
		cmdout["vfr"] = true;
		cmdout.removeMember("ifr");
	}
}

void SocketServer::cmd_crossing(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (!m_autoroute) {
		cmdout["error"] = "no autoroute instance";
		return;
	}
	if (cmdin.isMember("crossing")) {
		const Json::Value& crossing(cmdin["crossing"]);
		if (crossing.isArray()) {
			unsigned int index = 0;
			m_autoroute->set_crossing_size(0);
			for (unsigned int i = 0; i < crossing.size(); ++i) {
				const Json::Value& el(crossing[i]);
				if (!el.isObject())
					continue;
				{
					bool ok(false);
					Point pt(get_point(el, "coord"));
					if (pt.is_invalid()) {
						if (el.isMember("ident"))
							ok = m_autoroute->set_crossing(index, el.get("ident", "").asString());
					} else {
						m_autoroute->set_crossing(index, pt);
					        ok = true;
					}
					if (!ok)
						continue;
				}
				if (el.isMember("radius"))
					m_autoroute->set_crossing_radius(index, el["radius"].asDouble());
				if (el.isMember("minlevel") || el.isMember("maxlevel")) {
					int minlevel(m_autoroute->get_crossing_minlevel(index));
					int maxlevel(m_autoroute->get_crossing_maxlevel(index));
					if (el.isMember("minlevel"))
						minlevel = el["minlevel"].asInt();
					if (el.isMember("maxlevel"))
						maxlevel = el["maxlevel"].asInt();
					m_autoroute->set_crossing_level(index, minlevel, maxlevel);
				}
				++index;
			}
		}
	}
	Json::Value crossings;
	for (unsigned int i = 0; i < m_autoroute->get_crossing_size(); ++i) {
		Json::Value crossing;
		set_point(crossing, "coord", m_autoroute->get_crossing(i));
		crossing["type"] = to_str(m_autoroute->get_crossing_type(i));
		crossing["ident"] = m_autoroute->get_crossing_ident(i);
		crossing["radius"] = m_autoroute->get_crossing_radius(i);
		crossing["minlevel"] = m_autoroute->get_crossing_minlevel(i);
		crossing["maxlevel"] = m_autoroute->get_crossing_maxlevel(i);
		crossings.append(crossing);
	}
	cmdout["crossing"] = crossings;
}

void SocketServer::cmd_enroute(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (!m_autoroute) {
		cmdout["error"] = "no autoroute instance";
		return;
	}
	if (cmdin.isMember("dctlimit"))
		m_autoroute->set_dctlimit(cmdin["dctlimit"].asDouble());
	if (cmdin.isMember("dctpenalty"))
		m_autoroute->set_dctpenalty(cmdin["dctpenalty"].asDouble());
	if (cmdin.isMember("dctoffset"))
		m_autoroute->set_dctoffset(cmdin["dctoffset"].asDouble());
	if (cmdin.isMember("vfraspclimit"))
		m_autoroute->set_vfraspclimit(cmdin["vfraspclimit"].asDouble());
	if (cmdin.isMember("forceenrouteifr"))
		m_autoroute->set_force_enroute_ifr(!!cmdin["forceenrouteifr"].asBool());
	if (cmdin.isMember("honourawylevels"))
		m_autoroute->set_honour_awy_levels(!!cmdin["honourawylevels"].asBool());
	cmdout["dctlimit"] = m_autoroute->get_dctlimit();
	cmdout["dctpenalty"] = m_autoroute->get_dctpenalty();
	cmdout["dctoffset"] = m_autoroute->get_dctoffset();
	cmdout["vfraspclimit"] = m_autoroute->get_vfraspclimit();
	cmdout["forceenrouteifr"] = !!m_autoroute->get_force_enroute_ifr();
	cmdout["honourawylevels"] = !!m_autoroute->get_honour_awy_levels();
}

void SocketServer::cmd_levels(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (!m_autoroute) {
		cmdout["error"] = "no autoroute instance";
		return;
	}
	{
		int b(m_autoroute->get_baselevel());
		int t(m_autoroute->get_toplevel());
		if (cmdin.isMember("base"))
			b = cmdin["base"].asInt();
		if (cmdin.isMember("top"))
			t = cmdin["top"].asInt();
		m_autoroute->set_levels(b, t);
	}
	if (cmdin.isMember("maxdescent"))
		m_autoroute->set_maxdescent(cmdin["maxdescent"].asDouble());
	if (cmdin.isMember("honourlevelchangetrackmiles"))
		m_autoroute->set_honour_levelchangetrackmiles(!!cmdin["honourlevelchangetrackmiles"].asBool());
	cmdout["base"] = m_autoroute->get_baselevel();
	cmdout["top"] = m_autoroute->get_toplevel();
	cmdout["maxdescent"] = m_autoroute->get_maxdescent();
	cmdout["honourlevelchangetrackmiles"] = !!m_autoroute->get_honour_levelchangetrackmiles();
}

void SocketServer::cmd_exclude(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (!m_autoroute) {
		cmdout["error"] = "no autoroute instance";
		return;
	}
	if (cmdin.isMember("clear")) {
		const Json::Value& v(cmdin["clear"]);
		if (!v.isBool() || v.asBool())
			m_autoroute->clear_excluderegions();
	}
	bool indexf(cmdin.isMember("index"));
	unsigned int indexv(0);
	if (indexf)
		indexv = cmdin["index"].asUInt();
	{
		int ob(0), ot(999);
		double awylimit(0), dctlimit(0), dctoffset(0), dctscale(1);
		if (cmdin.isMember("base"))
			ob = cmdin["base"].asInt();
		if (cmdin.isMember("top"))
			ot = cmdin["top"].asInt();
		if (cmdin.isMember("awylimit"))
			awylimit = cmdin["awylimit"].asDouble();
		if (cmdin.isMember("dctlimit"))
			dctlimit = cmdin["dctlimit"].asDouble();
		if (cmdin.isMember("dctoffset"))
			dctoffset = cmdin["dctoffset"].asDouble();
		if (cmdin.isMember("dctscale"))
			dctscale = cmdin["dctscale"].asDouble();
		if (cmdin.isMember("aspcid")) {
			m_autoroute->add_excluderegion(CFMUAutoroute::ExcludeRegion(cmdin["aspcid"].asString(), cmdin.get("aspctype", "").asString(),
										    ob, ot, awylimit, dctlimit, dctoffset, dctscale));
			if (!indexf) {
				indexv = m_autoroute->get_excluderegions().size() - 1;
				indexf = true;
			}
		}
		Point ptsw(get_point(cmdin, "sw"));
		Point ptne(get_point(cmdin, "ne"));
		if (!ptsw.is_invalid() && !ptne.is_invalid()) {
			m_autoroute->add_excluderegion(CFMUAutoroute::ExcludeRegion(Rect(ptsw, ptne), ob, ot, awylimit, dctlimit, dctoffset, dctscale));
			if (!indexf) {
				indexv = m_autoroute->get_excluderegions().size() - 1;
				indexf = true;
			}
		}
	}
	cmdout["count"] = (Json::LargestUInt)m_autoroute->get_excluderegions().size();
	if (indexf && indexv < m_autoroute->get_excluderegions().size()) {
		const CFMUAutoroute::ExcludeRegion& er(m_autoroute->get_excluderegions()[indexv]);
		cmdout["base"] = er.get_minlevel();
		cmdout["top"] = er.get_maxlevel();
		cmdout["awylimit"] = er.get_awylimit();
		cmdout["dctlimit"] = er.get_dctlimit();
		cmdout["dctoffset"] = er.get_dctoffset();
		cmdout["dctscale"] = er.get_dctscale();
		if (er.is_airspace()) {
			cmdout["aspcid"] = er.get_airspace_id();
			cmdout["aspctype"] = er.get_airspace_type();
		} else {
			set_point(cmdout, "sw", er.get_bbox().get_southwest());
			set_point(cmdout, "ne", er.get_bbox().get_northeast());
		}
	}
}

void SocketServer::cmd_tfr(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (!m_autoroute) {
		cmdout["error"] = "no autoroute instance";
		return;
	}
	if (cmdin.isMember("enabled"))
		m_autoroute->set_tfr_enabled(!!cmdin["enabled"].asBool());
	if (cmdin.isMember("trace"))
		m_autoroute->set_tfr_trace(cmdin["trace"].asString());
	if (cmdin.isMember("disable"))
		m_autoroute->set_tfr_disable(cmdin["disable"].asString());
	if (cmdin.isMember("precompgraph"))
		m_autoroute->set_precompgraph_enabled(!!cmdin["precompgraph"].asBool());
	cmdout["enabled"] = !!m_autoroute->get_tfr_enabled();
	cmdout["available"] = !!m_autoroute->get_tfr_available();
	cmdout["trace"] = m_autoroute->get_tfr_trace();
	cmdout["disable"] = m_autoroute->get_tfr_disable();
	cmdout["precompgraph"] = !!m_autoroute->get_precompgraph_enabled();
}

void SocketServer::cmd_atmosphere(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (!m_autoroute) {
		cmdout["error"] = "no autoroute instance";
		return;
	}
	if (cmdin.isMember("qnh"))
		m_autoroute->set_qnh(cmdin["qnh"].asDouble());
	if (cmdin.isMember("isa"))
		m_autoroute->set_isaoffs(cmdin["isa"].asDouble());
	if (cmdin.isMember("wind"))
		m_autoroute->set_wind_enabled(!!cmdin["wind"].asBool());
	cmdout["qnh"] = m_autoroute->get_qnh();
	cmdout["isa"] = m_autoroute->get_isaoffs();
	cmdout["wind"] = !!m_autoroute->get_wind_enabled();
}

void SocketServer::cmd_cruise(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (!m_autoroute) {
		cmdout["error"] = "no autoroute instance";
		return;
	}
	if (cmdin.isMember("rpm") || cmdin.isMember("mp") || cmdin.isMember("bhp")) {
		double rpm(std::numeric_limits<double>::quiet_NaN());
		double mp(rpm), bhp(rpm);
		if (cmdin.isMember("rpm"))
			rpm = cmdin["rpm"].asDouble();
		if (cmdin.isMember("mp"))
			mp = cmdin["mp"].asDouble();
		if (cmdin.isMember("bhp"))
			bhp = cmdin["bhp"].asDouble();
		m_autoroute->set_engine_rpm(rpm);
		m_autoroute->set_engine_mp(mp);
     		m_autoroute->set_engine_bhp(bhp);
	}
	double rpm(m_autoroute->get_engine_rpm()), mp(m_autoroute->get_engine_mp()), bhp(m_autoroute->get_engine_bhp());
	if (!std::isnan(rpm))
		cmdout["rpm"] = rpm;
	if (!std::isnan(mp))
		cmdout["mp"] = mp;
	if (!std::isnan(bhp))
		cmdout["bhp"] = bhp;
}

void SocketServer::cmd_optimization(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (!m_autoroute) {
		cmdout["error"] = "no autoroute instance";
		return;
	}
	if (cmdin.isMember("target")) {
		std::string tgt(cmdin["target"].asString());
		static const CFMUAutoroute::opttarget_t ot[] = {
			CFMUAutoroute::opttarget_time,
			CFMUAutoroute::opttarget_fuel,
			CFMUAutoroute::opttarget_preferred
		};
		unsigned int i;
		for (i = 0; i < sizeof(ot)/sizeof(ot[0]); ++i)
			if (::to_str(ot[i]) == tgt)
				break;
		if (i >= sizeof(ot)/sizeof(ot[0])) {
			cmdout["error"] = "invalid target=" + tgt;
			return;
		}
		m_autoroute->set_opttarget(ot[i]);
	}
	cmdout["target"] = (std::string)to_str(m_autoroute->get_opttarget());
}

void SocketServer::cmd_preferred(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (!m_autoroute) {
		cmdout["error"] = "no autoroute instance";
		return;
	}
	if (cmdin.isMember("level"))
		m_autoroute->set_preferred_level(cmdin["level"].asInt());
	if (cmdin.isMember("penalty"))
		m_autoroute->set_preferred_penalty(cmdin["penalty"].asDouble());
	if (cmdin.isMember("climb"))
		m_autoroute->set_preferred_climb(cmdin["climb"].asDouble());
	if (cmdin.isMember("descent"))
		m_autoroute->set_preferred_descent(cmdin["descent"].asDouble());
	cmdout["level"] = m_autoroute->get_preferred_level();
	cmdout["penalty"] = m_autoroute->get_preferred_penalty();
	cmdout["climb"] = m_autoroute->get_preferred_climb();
	cmdout["descent"] = m_autoroute->get_preferred_descent();
}

void SocketServer::cmd_aircraft(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (!m_autoroute) {
		cmdout["error"] = "no autoroute instance";
		return;
	}
	bool dirty(false);
	if (cmdin.isMember("file")) {
		bool status(!!m_autoroute->get_aircraft().load_file(cmdin["file"].asString()));
		cmdout["status"] = status;
		dirty = status && (m_autoroute->get_aircraft().get_dist().recalculatepoly(false) ||
				   m_autoroute->get_aircraft().get_climb().recalculatepoly(m_autoroute->get_aircraft().get_vy(), false));
	}
	if (cmdin.isMember("data")) {
		bool status(!!m_autoroute->get_aircraft().load_string(cmdin["data"].asString()));
		cmdout["status"] = status;
		dirty = status && (m_autoroute->get_aircraft().get_dist().recalculatepoly(false) ||
				   m_autoroute->get_aircraft().get_climb().recalculatepoly(m_autoroute->get_aircraft().get_vy(), false));
	}
	if (cmdin.isMember("registration"))
		m_autoroute->get_aircraft().set_callsign(cmdin["registration"].asString());
	if (cmdin.isMember("type"))
		m_autoroute->get_aircraft().set_icaotype(cmdin["type"].asString());
	if (cmdin.isMember("transponder"))
		m_autoroute->get_aircraft().set_transponder(cmdin["transponder"].asString());
	{
		bool fixeq(false);
		if (cmdin.isMember("equipment")) {
			m_autoroute->get_aircraft().set_equipment(cmdin["equipment"].asString());
			fixeq = true;
		}
		if (cmdin.isMember("pbn")) {
			m_autoroute->get_aircraft().set_pbn(cmdin["pbn"].asString());
			fixeq = true;
		}
		if (fixeq)
			m_autoroute->get_aircraft().pbn_fix_equipment();
	}
	cmdout["registration"] = (std::string)m_autoroute->get_aircraft().get_callsign();
	cmdout["manufacturer"] = (std::string)m_autoroute->get_aircraft().get_manufacturer();
	cmdout["model"] = (std::string)m_autoroute->get_aircraft().get_model();
	cmdout["year"] = (std::string)m_autoroute->get_aircraft().get_year();
	cmdout["description"] = (std::string)m_autoroute->get_aircraft().get_description();
	cmdout["type"] = (std::string)m_autoroute->get_aircraft().get_icaotype();
	cmdout["equipment"] = (std::string)m_autoroute->get_aircraft().get_equipment();
	cmdout["transponder"] = (std::string)m_autoroute->get_aircraft().get_transponder();
	cmdout["pbn"] = (std::string)m_autoroute->get_aircraft().get_pbn_string();
	cmdout["data"] = (std::string)m_autoroute->get_aircraft().save_string();
	cmdout["dirty"] = dirty;
}

void SocketServer::cmd_start(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (!m_autoroute) {
		cmdout["error"] = "no autoroute instance";
		return;
	}
	m_autoroute->start();
}

void SocketServer::cmd_stop(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (!m_autoroute) {
		cmdout["error"] = "no autoroute instance";
		return;
	}
	m_autoroute->stop();
}

void SocketServer::cmd_continue(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (!m_autoroute) {
		cmdout["error"] = "no autoroute instance";
		return;
	}
	m_autoroute->cont();
}

void SocketServer::cmd_preload(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (!m_autoroute) {
		cmdout["error"] = "no autoroute instance";
		return;
	}
	m_autoroute->preload();
}

void SocketServer::cmd_clear(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (!m_autoroute) {
		cmdout["error"] = "no autoroute instance";
		return;
	}
	m_autoroute->clear();
}

void SocketServer::cmd_fplparse(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (!m_autoroute) {
		cmdout["error"] = "no autoroute instance";
		return;
	}
	if (!cmdin.isMember("fpl") || (!cmdin["fpl"].isString() && !cmdin["fpl"].isObject())) {
		cmdout["error"] = "no flight plan given";
		return;
	}
	FPlanRoute route(*(FPlan *)0);
	if (cmdin["fpl"].isString()) {
		std::string fpl(cmdin["fpl"].asString());
		cmdout["fpl"] = fpl;
		Engine eng(m_autoroute->get_db_maindir(), Engine::auxdb_override, m_autoroute->get_db_auxdir(), false, false);
		IcaoFlightPlan f(eng);
	        IcaoFlightPlan::errors_t err(f.parse(fpl, true));
		if (!err.empty()) {
			if (err.size() == 1) {
				cmdout["error"] = err[0];
			} else {
				Json::Value& er(cmdout["error"]);
				for (IcaoFlightPlan::errors_t::const_iterator i(err.begin()), e(err.end()); i != e; ++i)
					er.append(Json::Value(*i));
			}
		}
		f.set_route(route);
		cmdout["aircraftid"] = f.get_aircraftid();
		cmdout["flightrules"] = f.get_flightrules();
		cmdout["flighttype"] = f.get_flighttype();
		cmdout["number"] = f.get_number();
		cmdout["aircrafttype"] = f.get_aircrafttype();
		cmdout["wakecategory"] = f.get_wakecategory();
		cmdout["equipment"] = f.get_equipment();
		cmdout["transponder"] = f.get_transponder();
		cmdout["pbn"] = f.get_pbn_string();
		cmdout["cruisespeed"] = f.get_cruisespeed();
		cmdout["departure"] = f.get_departure();
		cmdout["destination"] = f.get_destination();
		cmdout["alternate1"] = f.get_alternate1();
		cmdout["alternate2"] = f.get_alternate2();
		cmdout["departuretime"] = (Json::LargestInt)f.get_departuretime();
		cmdout["totaleet"] = (Json::LargestInt)f.get_totaleet();
		cmdout["endurance"] = (Json::LargestInt)f.get_endurance();
		cmdout["personsonboard"] = f.get_personsonboard();
		cmdout["colourmarkings"] = f.get_colourmarkings();
		cmdout["remarks"] = f.get_remarks();
		cmdout["picname"] = f.get_picname();
		cmdout["emergencyradio"] = f.get_emergencyradio();
		cmdout["survival"] = f.get_survival();
		cmdout["lifejackets"] = f.get_lifejackets();
		cmdout["dinghies"] = f.get_dinghies();
	} else if (cmdin["fpl"].isObject()) {
		decode_fpl(route, cmdin["fpl"]);
	}
	route.turnpoints();
	route.set_winddir(0);
	route.set_windspeed(0);
	route.set_qff_hpa(m_autoroute->get_qnh());
	route.set_isaoffset_kelvin(m_autoroute->get_isaoffs());
	if (cmdin.isMember("recompute") && cmdin["recompute"].isBool() && cmdin["recompute"].asBool()) {
		route.recompute(m_autoroute->get_aircraft(), m_autoroute->get_qnh(), m_autoroute->get_isaoffs(),
				m_autoroute->get_engine_bhp(), m_autoroute->get_engine_rpm(), m_autoroute->get_engine_mp());
		
		bool gfs(cmdin.isMember("gfs") && cmdin["gfs"].isBool() && cmdin["gfs"].asBool());
		if (gfs) {
			Json::Value zw;
			encode_fpl(cmdout, route);
			FPlanRoute::GFSResult r(route.gfs(m_autoroute->get_grib2()));
			if (r.is_modified() && !r.is_partial()) {
				cmdout["zerowind"] = zw;
				route.recompute(m_autoroute->get_aircraft(), m_autoroute->get_qnh(), m_autoroute->get_isaoffs(),
						m_autoroute->get_engine_bhp(), m_autoroute->get_engine_rpm(), m_autoroute->get_engine_mp());
				cmdout["gfsminreftime"] = (std::string)Glib::TimeVal(r.get_minreftime(), 0).as_iso8601();
				cmdout["gfsmaxreftime"] = (std::string)Glib::TimeVal(r.get_maxreftime(), 0).as_iso8601();
				cmdout["gfsminefftime"] = (std::string)Glib::TimeVal(r.get_minefftime(), 0).as_iso8601();
				cmdout["gfsmaxefftime"] = (std::string)Glib::TimeVal(r.get_maxefftime(), 0).as_iso8601();
			} else {
				gfs = false;
				route.set_winddir(0);
				route.set_windspeed(0);
				route.set_qff_hpa(m_autoroute->get_qnh());
				route.set_isaoffset_kelvin(m_autoroute->get_isaoffs());
			}
		}
	}
	encode_fpl(cmdout, route);
	if (cmdin.isMember("localvalidate") && cmdin["localvalidate"].isBool() && cmdin["localvalidate"].asBool()) {
		Json::Value& localvalidate(cmdout["localvalidate"]);
		TrafficFlowRestrictions::Result res;
		bool chk(m_autoroute->check_fplan(res, route));
		if (chk) {
			{
				Json::Value& messages(localvalidate["messages"]);
				for (unsigned int i = 0, n = res.messages_size(); i < n; ++i) {
					const TrafficFlowRestrictions::Message& msg(res.get_message(i));
					Json::Value rmsg;
					Json::Value& vset(messages["vertices"]);
					for (TrafficFlowRestrictions::Message::set_t::const_iterator
						     i(msg.get_vertexset().begin()), e(msg.get_vertexset().end()); i != e; ++i)
						vset.append(Json::Value(*i));
					Json::Value& eset(messages["edges"]);
					for (TrafficFlowRestrictions::Message::set_t::const_iterator
						     i(msg.get_edgeset().begin()), e(msg.get_edgeset().end()); i != e; ++i)
						eset.append(Json::Value(*i));
					rmsg["text"] = msg.get_text();
					rmsg["rule"] = msg.get_rule();
					rmsg["type"] = msg.get_type_string();
				}
			}
			{
				Json::Value& rules(localvalidate["rules"]);
				for (unsigned int i = 0, n = res.size(); i < n; ++i) {
					Json::Value rrule;
					encode_rule(rrule, res[i]);
					rules.append(rrule);
				}
			}
			{
				const TrafficFlowRestrictions::Fpl& fpl(res.get_fplan());
				Json::Value& expfpl(localvalidate["expandedfpl"]);
				for (unsigned int i = 0; i < fpl.size(); ++i) {
					const TrafficFlowRestrictions::FplWpt& wpt(fpl[i]);
					Json::Value fplwpt;
					encode_wpt(fplwpt, wpt);
					fplwpt["wptnr"] = wpt.get_wptnr();
					fplwpt["ndb"] = wpt.has_ndb();
					fplwpt["vor"] = wpt.has_vor();
					fplwpt["dme"] = wpt.has_dme();
				        fplwpt["ident"] = (std::string)wpt.get_ident();
					expfpl.append(fplwpt);
				}
			}
			localvalidate["aircrafttype"] = res.get_aircrafttype();
			localvalidate["aircrafttypeclass"] = res.get_aircrafttypeclass();
			localvalidate["equipment"] = res.get_equipment();
			localvalidate["pbn"] = res.get_pbn_string();
			localvalidate["typeofflight"] = res.get_typeofflight();
		} else {
			localvalidate["error"] = "TFR disabled";
		}
	}
	{
		Engine eng(m_autoroute->get_db_maindir(), Engine::auxdb_override, m_autoroute->get_db_auxdir(), false, false);
		IcaoFlightPlan f(eng);
		f.set_aircraft(m_autoroute->get_aircraft(), m_autoroute->get_engine_rpm(), m_autoroute->get_engine_mp(),
			       m_autoroute->get_engine_bhp());
		f.set_personsonboard(0);
		f.populate(route, IcaoFlightPlan::awymode_keep, 50.0);
		cmdout["fplall"] = f.get_fpl();
		if (!cmdin["fpl"].isString())
			cmdout["fpl"] = f.get_fpl();
		f.populate(route, IcaoFlightPlan::awymode_collapse, 50.0);
		cmdout["fplawy"] = f.get_fpl();
		f.populate(route, IcaoFlightPlan::awymode_collapse_dct, 50.0);
		cmdout["fplawydct"] = f.get_fpl();
		f.populate(route, IcaoFlightPlan::awymode_collapse_all, 50.0);
		cmdout["fplmin"] = f.get_fpl();
	}
}

Json::Value SocketServer::servercmd(const Json::Value& cmdin)
{
	std::string cmdname(cmdin.get("cmdname", "").asString());
	cmdlist_t::const_iterator cmdi(m_servercmdlist.find(cmdname));
	if (cmdi == m_servercmdlist.end())
		return Json::Value();
	Json::Value reply;
	if (cmdin.isMember("cmdseq"))
		reply["cmdseq"] = cmdin["cmdseq"];
	reply["cmdname"] = cmdname;
	(this->*(cmdi->second))(cmdin, reply);
	return reply;
}

void SocketServer::servercmd_scangrib2(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (!m_autoroute) {
		cmdout["error"] = "no autoroute instance";
		return;
	}
	typedef std::vector<std::string> pathlist_t;
	pathlist_t pathlist;
	if (cmdin.isMember("path")) {
		const Json::Value& paths(cmdin["path"]);
		if (paths.isString()) {
			pathlist.push_back(paths.asString());
		} else if (paths.isArray()) {
			for (Json::ArrayIndex i = 0; i < paths.size(); ++i) {
				const Json::Value& path(paths[i]);
				if (!path.isString())
					continue;
				pathlist.push_back(path.asString());
			}
		}
	}
	if (pathlist.empty()) {
		std::string path(m_autoroute->get_db_maindir());
		if (path.empty())
			path = FPlan::get_userdbpath();
		path = Glib::build_filename(path, "gfs");
		pathlist.push_back(path);
		if (!m_autoroute->get_db_auxdir().empty())
			pathlist.push_back(m_autoroute->get_db_auxdir());
	}
	Json::Value& paths(cmdout["paths"]);
	GRIB2& grib2(m_autoroute->get_grib2());
	GRIB2::Parser p(grib2);
	unsigned int count(grib2.find_layers().size());
	for (pathlist_t::const_iterator pi(pathlist.begin()), pe(pathlist.end()); pi != pe; ++pi) {
		p.parse(*pi);
		unsigned int count1(grib2.find_layers().size());
		Json::Value path;
		path["path"] = *pi;
		path["count"] = count1 - count;
		count = count1;
		paths.append(path);
	}
	cmdout["obsolete"] = grib2.remove_obsolete_layers();
}

void SocketServer::servercmd_processes(const Json::Value& cmdin, Json::Value& cmdout)
{
	Json::Value& rtrs(cmdout["router"]);
	{
		Glib::TimeVal tv;
		tv.assign_current_time();
		cmdout["time"] = (std::string)tv.as_iso8601();
	}
	for (routers_t::const_iterator ri(m_routers.begin()), re(m_routers.end()); ri != re; ++ri) {
		if (!ri->second)
			continue;
		Json::Value rtr;
		rtr["rxqueuesize"] = ri->second->get_recvqueue_size();
		rtr["accesstime"] = (std::string)ri->second->get_accesstime().as_iso8601();
		rtr["starttime"] = (std::string)ri->second->get_starttime().as_iso8601();
		rtr["session"] = ri->second->get_session();
		rtr["pid"] = ri->second->get_pid();
		rtr["running"] = ri->second->is_running();
		rtr["dead"] = ri->second->is_dead();
		rtrs.append(rtr);
	}
}

void SocketServer::servercmd_ruleinfo(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (!m_autoroute) {
		cmdout["error"] = "no autoroute instance";
		return;
	}
	cmdout["nrrules"] = m_autoroute->get_tfr().count_rules();
	cmdout["nrdctrules"] = m_autoroute->get_tfr().count_dct_rules();
	bool exact(!cmdin.isMember("exact") || !cmdin["exact"].isBool() || cmdin["exact"].asBool());
	std::vector<TrafficFlowRestrictions::RuleResult> res;
	if (cmdin.isMember("rules")) {
		const Json::Value& rules(cmdin["rules"]);
		if (rules.isString()) {
			res = m_autoroute->get_tfr().find_rules(rules.asString(), exact);
		} else if (rules.isArray()) {
			for (unsigned int i = 0, n = rules.size(); i < n; ++i) {
				const Json::Value& rule(rules[i]);
				if (!rule.isString())
					continue;
				std::vector<TrafficFlowRestrictions::RuleResult> res1(m_autoroute->get_tfr().find_rules(rule.asString(), exact));
				res.insert(res.end(), res1.begin(), res1.end());
			}
		}
	}
	{
		Json::Value& rules(cmdout["rules"]);
		for (std::vector<TrafficFlowRestrictions::RuleResult>::const_iterator i(res.begin()), e(res.end()); i != e; ++i) {
			Json::Value rule;
			encode_rule(rule, *i);
			rules.append(rule);
		}
	}
}
