#ifndef CFMUAUTOROUTE_H
#define CFMUAUTOROUTE_H

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "sysdeps.h"
#include "fplan.h"
#include "dbobj.h"
#include "engine.h"
#include "aircraft.h"
#include "baro.h"
#include "wind.h"
#include "tfr.hh"

#include <boost/config.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <glibmm.h>
#include <giomm.h>
#include <sigc++/sigc++.h>

// TODO:
// Path Probing
// Implement Area deletion (for EDGGADF)
// Implement per country DCT limit
// Implement SID/STAR for VFR and IFR

class CFMUAutoroute : public sigc::trackable {
public:
	CFMUAutoroute();
	~CFMUAutoroute();

	void set_db_directories(const std::string& dir_main = "", Engine::auxdb_mode_t auxdbmode = Engine::auxdb_prefs, const std::string& dir_aux = "");
	const std::string& get_db_maindir(void) const { return m_dbdir_main; }
	const std::string& get_db_auxdir(void) const { return m_dbdir_aux; }
	Preferences& get_prefs(void) { return m_prefs; }

	Aircraft& get_aircraft(void) { return m_aircraft; }
	const Aircraft& get_aircraft(void) const { return m_aircraft; }

	GRIB2& get_grib2(void) { return m_grib2; }
	const GRIB2& get_grib2(void) const { return m_grib2; }

	TrafficFlowRestrictions& get_tfr(void) { return m_tfr; }
	const TrafficFlowRestrictions& get_tfr(void) const { return m_tfr; }

	bool check_fplan(TrafficFlowRestrictions::Result& res, const FPlanRoute& route);

	typedef enum {
		opttarget_time,
		opttarget_fuel,
		opttarget_preferred
	} opttarget_t;
	void set_opttarget(opttarget_t t = opttarget_time);
	opttarget_t get_opttarget(void) const { return m_opttarget; }

	const AirportsDb::Airport& get_departure(void) const { return m_airport[0]; }
	void set_departure(const AirportsDb::Airport& el);
	bool set_departure(const std::string& icao, const std::string& name);
	bool get_departure_ifr(void) const { return m_airportifr[0]; }
	void set_departure_ifr(bool ifr);
	const Point& get_sid(void) const { return m_airportconn[0]; }
	void set_sid(const Point& pt = Point::invalid);
	bool set_sid(const std::string& name);
	Engine::AirwayGraphResult::Vertex::type_t get_sidtype(void) const { return m_airportconntype[0]; }
	const std::string& get_sidident(void) const { return m_airportconnident[0]; }
	double get_sidlimit(void) const { return m_airportconnlimit[0]; }
	void set_sidlimit(double l = 20.0);
	double get_sidpenalty(void) const { return m_airportconnpenalty[0]; }
	void set_sidpenalty(double p = 1.0);
	double get_sidoffset(void) const { return m_airportconnoffset[0]; }
	void set_sidoffset(double o = 0.0);
	double get_sidminimum(void) const { return m_airportconnminimum[0]; }
	void set_sidminimum(double m = 0.0);
	bool get_siddb(void) const { return m_airportconndb[0]; }
	void set_siddb(bool db);

	const AirportsDb::Airport& get_destination(void) const { return m_airport[1]; }
	void set_destination(const AirportsDb::Airport& el);
	bool set_destination(const std::string& icao, const std::string& name);
	bool get_destination_ifr(void) const { return m_airportifr[1]; }
	void set_destination_ifr(bool ifr);
	const Point& get_star(void) const { return m_airportconn[1]; }
	void set_star(const Point& pt = Point::invalid);
	bool set_star(const std::string& name);
	Engine::AirwayGraphResult::Vertex::type_t get_startype(void) const { return m_airportconntype[1]; }
	const std::string& get_starident(void) const { return m_airportconnident[1]; }
	double get_starlimit(void) const { return m_airportconnlimit[1]; }
	void set_starlimit(double l = 20.0);
	double get_starpenalty(void) const { return m_airportconnpenalty[1]; }
	void set_starpenalty(double p = 1.0);
	double get_staroffset(void) const { return m_airportconnoffset[1]; }
	void set_staroffset(double o = 0.0);
	double get_starminimum(void) const { return m_airportconnminimum[1]; }
	void set_starminimum(double m = 0.0);
	bool get_stardb(void) const { return m_airportconndb[1]; }
	void set_stardb(bool db);

	class Crossing {
	public:
		static const double maxradius = 50;
		Crossing() : m_radius(0), m_minlevel(0), m_maxlevel(600), m_type(Engine::AirwayGraphResult::Vertex::type_undefined) { m_coord.set_invalid(); }
		const std::string& get_ident(void) const { return m_ident; }
		void set_ident(const std::string& x = "") { m_ident = x; }
		const Point& get_coord(void) const { return m_coord; }
		void set_coord(const Point& pt) { m_coord = pt; }
		void set_coord(void) { m_coord.set_invalid(); }
		double get_radius(void) const { return m_radius; }
		void set_radius(double r = 0) { m_radius = std::max(std::min(r, maxradius), 0.0); }
		int get_minlevel(void) const { return m_minlevel; }
		int get_maxlevel(void) const { return m_maxlevel; }
		void set_level(int minlevel = 0, int maxlevel = 600) {
			m_minlevel = minlevel;
			m_maxlevel = maxlevel;
			if (m_maxlevel < m_minlevel)
				std::swap(m_maxlevel, m_minlevel);
		}
		Engine::AirwayGraphResult::Vertex::type_t get_type(void) const { return m_type; }
		void set_type(Engine::AirwayGraphResult::Vertex::type_t typ = Engine::AirwayGraphResult::Vertex::type_undefined) { m_type = typ; }

	protected:
		std::string m_ident;
		Point m_coord;
		double m_radius;
		int m_minlevel;
		int m_maxlevel;
		Engine::AirwayGraphResult::Vertex::type_t m_type;
	};
	typedef std::vector<Crossing> crossing_t;

	void set_crossing(unsigned int index, const Point& pt);
	bool set_crossing(unsigned int index, const std::string& ident);
	void set_crossing_radius(unsigned int index, double r = 0);
	void set_crossing_level(unsigned int index, int minlevel = 0, int maxlevel = 600);
	const Point& get_crossing(unsigned int index) const;
	const std::string& get_crossing_ident(unsigned int index) const;
	double get_crossing_radius(unsigned int index) const;
	int get_crossing_minlevel(unsigned int index) const;
	int get_crossing_maxlevel(unsigned int index) const;
	Engine::AirwayGraphResult::Vertex::type_t get_crossing_type(unsigned int index) const;
	unsigned int get_crossing_size(void) const { return m_crossing.size(); }
	void set_crossing_size(unsigned int sz);

	double get_dctlimit(void) const { return m_dctlimit; }
	void set_dctlimit(double l);
	double get_dctpenalty(void) const { return m_dctpenalty; }
	void set_dctpenalty(double l);
	double get_dctoffset(void) const { return m_dctoffset; }
	void set_dctoffset(double l);
	double get_vfraspclimit(void) const { return m_vfraspclimit; }
	void set_vfraspclimit(double l);

	int get_baselevel(void) const { return m_levels[0]; }
	int get_toplevel(void) const { return m_levels[1]; }
	void set_levels(int base, int top);
	double get_maxdescent(void) const { return m_maxdescent; }
	void set_maxdescent(double d);
	bool get_honour_levelchangetrackmiles(void) const { return m_honourlevelchangetrackmiles; }
	void set_honour_levelchangetrackmiles(bool lvltrk);

	bool get_honour_awy_levels(void) const { return m_honourawylevels; }
	void set_honour_awy_levels(bool awylvl);

	void set_preferred_level(int level);
	int get_preferred_level(void) const { return m_preferredlevel; }
	void set_preferred_penalty(double p);
	double get_preferred_penalty(void) const { return m_preferredpenalty; }
	void set_preferred_climb(double p);
	double get_preferred_climb(void) const { return m_preferredclimb; }
	void set_preferred_descent(double p);
	double get_preferred_descent(void) const { return m_preferreddescent; }

	class ExcludeRegion {
	public:
		ExcludeRegion(const std::string& aspcid, const std::string& aspctype, int minlevel = 0, int maxlevel = 999,
			      double awylimit = 0, double dctlimit = 0, double dctoffset = 0, double dctscale = 1)
			: m_aspcid(aspcid), m_aspctype(aspctype), m_minlevel(minlevel), m_maxlevel(maxlevel),
			  m_awylimit(awylimit), m_dctlimit(dctlimit), m_dctoffset(dctoffset), m_dctscale(dctscale) {}
		ExcludeRegion(const Rect& bbox, int minlevel = 0, int maxlevel = 999,
			      double awylimit = 0, double dctlimit = 0, double dctoffset = 0, double dctscale = 1)
			: m_bbox(bbox), m_minlevel(minlevel), m_maxlevel(maxlevel),
			  m_awylimit(awylimit), m_dctlimit(dctlimit), m_dctoffset(dctoffset), m_dctscale(dctscale) {}
		bool is_airspace(void) const { return !(m_aspcid.empty() && m_aspctype.empty()); }
		const std::string& get_airspace_id(void) const { return m_aspcid; }
		const std::string& get_airspace_type(void) const { return m_aspctype; }
		const Rect& get_bbox(void) const { return m_bbox; }
		int get_minlevel(void) const { return m_minlevel; }
		int get_maxlevel(void) const { return m_maxlevel; }
		double get_awylimit(void) const { return m_awylimit; }
		double get_dctlimit(void) const { return m_dctlimit; }
		double get_dctoffset(void) const { return m_dctoffset; }
		double get_dctscale(void) const { return m_dctscale; }

	protected:
		std::string m_aspcid;
		std::string m_aspctype;
		Rect m_bbox;
		int m_minlevel;
		int m_maxlevel;
		double m_awylimit;
		double m_dctlimit;
		double m_dctoffset;
		double m_dctscale;
	};
	typedef std::vector<ExcludeRegion> excluderegions_t;

	const excluderegions_t& get_excluderegions(void) const { return m_excluderegions; }
	void set_excluderegions(const excluderegions_t& r);
	void add_excluderegion(const ExcludeRegion& r);
	void clear_excluderegions(void);

	bool get_force_enroute_ifr(void) const { return m_forceenrifr; }
	void set_force_enroute_ifr(bool ifr);

	bool get_tfr_available(void) const { return m_tfravailable; }
	bool get_tfr_enabled(void) const { return m_tfrenable; }
	void set_tfr_enabled(bool ena = true);

	void set_tfr_trace(const std::string& rules);
	std::string get_tfr_trace(void) const;
	void set_tfr_disable(const std::string& rules);
	std::string get_tfr_disable(void) const;
	void set_tfr_savefile(const std::string& fn) { m_tfrsavefile = fn; }
	std::string get_tfr_savefile(void) const { return m_tfrsavefile; }

	bool get_precompgraph_enabled(void) const { return m_precompgraphenable; }
	void set_precompgraph_enabled(bool ena = true);

	bool get_wind_enabled(void) const { return m_windenable; }
	void set_wind_enabled(bool ena = true);

	double get_qnh(void) const { return m_qnh; }
	void set_qnh(double q = IcaoAtmosphere::std_sealevel_pressure);
	double get_isaoffs(void) const { return m_isaoffs; }
	void set_isaoffs(double t = 0);
	double get_engine_rpm(void) const { return m_enginerpm; }
	void set_engine_rpm(double l);
	double get_engine_mp(void) const { return m_enginemp; }
	void set_engine_mp(double l);
	double get_engine_bhp(void) const { return m_enginebhp; }
	void set_engine_bhp(double l);

	time_t get_deptime(void) const { return m_route.get_time_offblock_unix(); }
	void set_deptime(time_t t);

	typedef enum {
		validator_default,
		validator_cfmu,
		validator_eurofpl
	} validator_t;

	validator_t get_validator(void) const { return m_validator; }
	void set_validator(validator_t v = validator_cfmu);
	std::string get_validator_socket(void) const;
	void set_validator_socket(const std::string& path = "");
	const std::string& get_validator_binary(void) const { return m_childbinary; }
	void set_validator_binary(const std::string& path = "");
	int get_validator_xdisplay(void) const { return m_childxdisplay; }
	void set_validator_xdisplay(int xdisp = -1);

	const FPlanRoute& get_route(void) const { return m_route; }
	std::string get_simplefpl(void) const;
	typedef std::vector<std::string> validationresponse_t;
	const validationresponse_t& get_validationresponse(void) const { return m_validationresponse; }

	double get_gcdistance(void) const;
	double get_routedistance(void) const;
	unsigned int get_mintime(bool wind = true) const;
	unsigned int get_routetime(bool wind = true) const;
	double get_minfuel(bool wind = true) const;
	double get_routefuel(bool wind = true) const;
	unsigned int get_local_iterationcount(void) const { return m_iterationcount[0]; }
	unsigned int get_remote_iterationcount(void) const { return m_iterationcount[1]; }
	unsigned int get_iterationcount(void) const { return get_local_iterationcount() + get_remote_iterationcount(); }

	void get_cruise_performance(int& alt, double& densityalt, double& tas) const;

	bool is_running(void) const;
	bool is_errorfree(void) const;

	Glib::TimeVal get_wallclocktime(void) const;
	Glib::TimeVal get_validatortime(void) const;

	void preload(bool validator = true);
	void start(void);
	void cont(void);
	void stop(void);
	void clear(void);

	void precompute_graph(const Rect& bbox, const std::string& fn);

	typedef enum {
		statusmask_starting = 1U << 0,
		statusmask_stoppingdone = 1U << 1,
		statusmask_stoppingerror = 1U << 2,
		statusmask_newfpl = 1U << 3,
		statusmask_newvalidateresponse = 1U << 4
	} statusmask_t;
	sigc::signal<void,statusmask_t> signal_statuschange(void) { return m_signal_statuschange; }

	typedef enum {
		log_fplproposal,
		log_fplvalidation,
		log_graphrule,
		log_graphruledesc,
		log_graphruleoprgoal,
		log_graphchange,
		log_precompgraph,
		log_weatherdata,
		log_normal,
		log_debug0,
		log_debug1
	} log_t;
	sigc::signal<void,log_t,const std::string&> signal_log(void) { return m_signal_log; }

protected:
	typedef bool (CFMUAutoroute::*parseresponse_t)(Glib::MatchInfo&);
	static const struct parsefunctions {
		const char *regex;
		parseresponse_t func;
	} parsefunctions[];
	static const char *ignoreregex[];

	class Performance {
	public:
		class Cruise {
		public:
			Cruise(int alt = 0, double da = 0, double ta = 0, double rpm = 0, double mp = 0,
			       double secpernmi = 0, double fuelpersec = 0, double metricpernmi = 0,
			       Glib::RefPtr<GRIB2::LayerResult> windu = Glib::RefPtr<GRIB2::LayerResult>(),
			       Glib::RefPtr<GRIB2::LayerResult> windv = Glib::RefPtr<GRIB2::LayerResult>(),
			       Glib::RefPtr<GRIB2::LayerResult> temp = Glib::RefPtr<GRIB2::LayerResult>());

			int get_altitude(void) const { return m_altitude; }
			double get_densityaltitude(void) const { return m_densityaltitude; }
			double get_truealtitude(void) const { return m_truealt; }
			static std::string get_fplalt(int alt);
			std::string get_fplalt(void) const { return get_fplalt(get_altitude()); }

			// Cruise Performance
			double get_secpernmi(void) const { return m_secpernmi; }
			double get_fuelpersec(void) const { return m_fuelpersec; }
			double get_metricpernmi(void) const { return m_metricpernmi; }

			// For FPL
			double get_tas(void) const { return 3600.0 / m_secpernmi; }
			double get_inverse_tas(void) const { return (1.0 / 3600.0) * m_secpernmi; }
			static std::string get_fpltas(double tas);
			std::string get_fpltas(void) const { return get_fpltas(get_tas()); }

			// Wind
			bool is_wind(void) const { return m_windu && m_windv; }
			Wind<double> get_wind(const Point& pt) const;
			bool is_temp(void) const { return m_temp; }
			double get_temp(const Point& pt) const;

			// Engine
			double get_rpm(void) const { return m_rpm; }
			double get_mp(void) const { return m_mp; }

		protected:
			Glib::RefPtr<GRIB2::LayerResult> m_windu;
			Glib::RefPtr<GRIB2::LayerResult> m_windv;
			Glib::RefPtr<GRIB2::LayerResult> m_temp;
			double m_secpernmi;
			double m_fuelpersec;
			double m_metricpernmi;
			double m_rpm;
			double m_mp;
			double m_densityaltitude;
			double m_truealt;
			int m_altitude;
		};

		class LevelChange {
		public:
			LevelChange(double tracknmi = 0, double timepenalty = 0, double fuelpenalty = 0, double metricpenalty = 0);
			double get_tracknmi(void) const { return m_tracknmi; }
			double get_timepenalty(void) const { return m_timepenalty; }
			double get_fuelpenalty(void) const { return m_fuelpenalty; }
			double get_metricpenalty(void) const { return m_metricpenalty; }

		protected:
			double m_tracknmi;
			double m_timepenalty;
			double m_fuelpenalty;
			double m_metricpenalty;
		};

		Performance(void);
		bool empty(void) const { return m_cruise.empty(); }
		unsigned int size(void) const { return m_cruise.size(); }
		const Cruise& get_cruise(unsigned int pi) const;
		const Cruise& get_cruise(unsigned int piu, unsigned int piv) const;
		const LevelChange& get_levelchange(unsigned int piu, unsigned int piv) const;
		void clear(void);
		void add(const Cruise& x);
		void set(unsigned int piu, unsigned int piv, const LevelChange& lvlchg);
		void resize(unsigned int pis);
		unsigned int findcruiseindex(int alt) const;
		int get_minaltitude(void) const;
		int get_maxaltitude(void) const;
		const Cruise& get_lowest_cruise(void) const;
		const Cruise& get_highest_cruise(void) const;

	protected:
		typedef std::vector<Cruise> cruise_t;
		cruise_t m_cruise;
		typedef std::vector<std::vector<LevelChange> > levelchange_t;
		levelchange_t m_levelchange;
		static const Cruise nullcruise;
	};

        Preferences m_prefs;
	AirportsDb m_airportdb;
	NavaidsDb m_navaiddb;
	WaypointsDb m_waypointdb;
	AirwaysDb m_airwaydb;
	AirspacesDb m_airspacedb;
	MapelementsDb m_mapelementdb;
	TopoDb30 m_topodb;
	GRIB2 m_grib2;
	FPlan m_fplandb;
	TrafficFlowRestrictions m_tfr;
	typedef std::set<std::string> ruleset_t;
	ruleset_t m_systemdisabledrules;
	ruleset_t m_systemtracerules;
	ruleset_t m_disabledrules;
	ruleset_t m_tracerules;
	std::string m_tfrsavefile;
	std::string m_dbdir_main;
	std::string m_dbdir_aux;
	Aircraft m_aircraft;
	FPlanRoute m_route;
	sigc::signal<void,statusmask_t> m_signal_statuschange;
	sigc::signal<void,log_t,const std::string&> m_signal_log;
	Performance m_performance;
	Glib::RefPtr<GRIB2::LayerResult> m_grib2prmsl;
	unsigned int m_routetime;
	unsigned int m_routezerowindtime;
	double m_routefuel;
	double m_routezerowindfuel;
	AirportsDb::Airport m_airport[2];
	Point m_airportconn[2];
	std::string m_airportconnident[2];
	double m_airportconnlimit[2];
	double m_airportconnpenalty[2];
	double m_airportconnoffset[2];
	double m_airportconnminimum[2];
	Engine::AirwayGraphResult::Vertex::type_t m_airportconntype[2];
	bool m_airportconndb[2];
	bool m_airportifr[2];
	crossing_t m_crossing;
	double m_maxdescent;
	double m_dctlimit;
	double m_dctpenalty;
	double m_dctoffset;
	double m_vfraspclimit;
	double m_qnh;
	double m_isaoffs;
	double m_enginerpm;
	double m_enginemp;
	double m_enginebhp;
	int m_levels[2];
	time_t m_starttime;
	int m_preferredlevel;
	double m_preferredpenalty;
	double m_preferredclimb;
	double m_preferreddescent;
	excluderegions_t m_excluderegions;
	std::string m_callsign;
	validationresponse_t m_validationresponse;
	Glib::RefPtr<Gio::Socket> m_childsock;
#ifdef G_OS_UNIX
	Glib::RefPtr<Gio::UnixSocketAddress> m_childsockaddr;
#else
	Glib::RefPtr<Gio::InetSocketAddress> m_childsockaddr;
#endif
	std::string m_childstdout;
	std::string m_childbinary;
	sigc::connection m_connchildtimeout;
	sigc::connection m_connchildwatch;
	sigc::connection m_connchildstdout;
	Glib::RefPtr<Glib::IOChannel> m_childchanstdin;
	Glib::RefPtr<Glib::IOChannel> m_childchanstdout;
	Glib::Pid m_childpid;
	unsigned int m_childtimeoutcount;
	int m_childxdisplay;
	unsigned int m_iterationcount[2];
	int m_pathprobe;
	opttarget_t m_opttarget;
	validator_t m_validator;
	Glib::TimeVal m_tvelapsed;
	Glib::TimeVal m_tvvalidator;
	Glib::TimeVal m_tvvalidatestart;
	bool m_childrun;
	bool m_forceenrifr;
	bool m_tfravailable;
	bool m_tfrenable;
	bool m_tfrloaded;
	bool m_honourawylevels;
	bool m_honourlevelchangetrackmiles;
	bool m_precompgraphenable;
	bool m_windenable;
	bool m_grib2loaded;
	bool m_running;
	bool m_done;

	template <typename T> static bool is_nan_or_inf(T x) { return std::isnan(x) || std::isinf(x); }
	void opendb(void);
	void closedb(void);
	std::string get_tfr_xml_filename(void) const { return Glib::build_filename(get_db_auxdir(), "eRAD.xml"); }
	std::string get_tfr_bin_filename(void) const { return Glib::build_filename(get_db_auxdir(), "eRAD.bin"); }
	std::string get_tfr_local_xml_filename(void) const { return Glib::build_filename(get_db_auxdir(), "eRAD_local.xml"); }
	std::string get_tfr_local_bin_filename(void) const { return Glib::build_filename(get_db_auxdir(), "eRAD_local.bin"); }
	std::string get_tfr_user_xml_filename(void) const { return Glib::build_filename(FPlan::get_userdbpath(), "eRAD.xml"); }
	std::string get_tfr_user_bin_filename(void) const { return Glib::build_filename(FPlan::get_userdbpath(), "eRAD.bin"); }
	std::string get_tfr_global_rules_filename(void) const { return Glib::build_filename(get_db_auxdir(), "eRAD_rules.xml"); }
	std::string get_tfr_local_rules_filename(void) const { return Glib::build_filename(FPlan::get_userdbpath(), "eRAD_rules.xml"); }
	void check_tfr_available(void);
	bool find_airport(AirportsDb::Airport& el, const std::string& icao, const std::string& name);
	bool find_point(Point& pt, std::string& ident, Engine::AirwayGraphResult::Vertex::type_t& typ, const std::string& name, bool ifr, const Point& ptnear, bool exact);
	bool is_optfuel(void) const { return m_opttarget == opttarget_fuel; }
	bool is_optpreferred(void) const { return m_opttarget == opttarget_preferred; }
	void updatefpl(void);
	Rect get_bbox(void) const;
	void logmessage(const TrafficFlowRestrictions::Message& msg);
	void loadtfr(void);
	void tfrpreparedct(void);
	void logwxlayers(const GRIB2::layerlist_t& ll, int alt = -1);
	void computeperformance(void);
	unsigned int findperfindex(int alt) const { return m_performance.findcruiseindex(alt); }
	void start(bool cont);
	bool start_vfr(void);
	bool start_ifr(bool cont = false, bool optfuel = false);
	bool start_ifr_specialcase(void);
	bool start_ifr_specialcase_pogo(void);
	void preload_wind(void);
	bool is_stopped(void);
	void child_watch(GPid pid, int child_status);
	void child_run(void);
	void child_close(void);
	bool child_stdout_handler(Glib::IOCondition iocond);
	bool child_socket_handler(Glib::IOCondition iocond);
	bool child_timeout(void);
	void child_send(const std::string& tx);
	bool child_is_running(void) const;
	void child_configure(void);
	unsigned int child_get_timeout(void) const { return m_childrun ? 120 : 30; }
	void parse_response(void);
	bool parse_response_route49(Glib::MatchInfo& minfo);
	bool parse_response_route52(Glib::MatchInfo& minfo);
	bool parse_response_route130(Glib::MatchInfo& minfo);
	bool parse_response_route134(Glib::MatchInfo& minfo);
	bool parse_response_route135(Glib::MatchInfo& minfo);
	bool parse_response_route139(Glib::MatchInfo& minfo);
	bool parse_response_route140(Glib::MatchInfo& minfo);
	bool parse_response_route165(Glib::MatchInfo& minfo);
	bool parse_response_route168(Glib::MatchInfo& minfo);
	bool parse_response_route171(Glib::MatchInfo& minfo);
	bool parse_response_route172(Glib::MatchInfo& minfo);
	bool parse_response_route179(Glib::MatchInfo& minfo);
	bool parse_response_prof50(Glib::MatchInfo& minfo);
	bool parse_response_prof193(Glib::MatchInfo& minfo);
	bool parse_response_prof194(Glib::MatchInfo& minfo);
	bool parse_response_prof195(Glib::MatchInfo& minfo);
	bool parse_response_prof195b(Glib::MatchInfo& minfo);
	bool parse_response_prof197(Glib::MatchInfo& minfo);
	bool parse_response_prof198(Glib::MatchInfo& minfo);
	bool parse_response_prof199(Glib::MatchInfo& minfo);
	bool parse_response_prof201(Glib::MatchInfo& minfo);
	bool parse_response_prof201b(Glib::MatchInfo& minfo);
	bool parse_response_prof204(Glib::MatchInfo& minfo);
	bool parse_response_prof204b(Glib::MatchInfo& minfo);
	bool parse_response_prof204c(Glib::MatchInfo& minfo);
	bool parse_response_prof204d(Glib::MatchInfo& minfo);
	bool parse_response_prof204e(Glib::MatchInfo& minfo);
	bool parse_response_prof205(Glib::MatchInfo& minfo);
	bool parse_response_prof205b(Glib::MatchInfo& minfo);
	bool parse_response_prof205c(Glib::MatchInfo& minfo);
	bool parse_response_prof206(Glib::MatchInfo& minfo);
	bool parse_response_efpm228(Glib::MatchInfo& minfo);
	bool parse_response_fail(Glib::MatchInfo& minfo);

	class VFRRouter {
	public:
		typedef Engine::AirwayGraphResult::Graph Graph;
		typedef Engine::AirwayGraphResult::Vertex Vertex;
		typedef Engine::AirwayGraphResult::Edge Edge;

		VFRRouter(CFMUAutoroute& ar);
		bool load_graph(const Rect& bbox);
		bool set_departure(const AirportsDb::Airport& arpt);
		bool set_destination(const AirportsDb::Airport& arpt);
		void add_dct(double dctlim, double sidlim, double starlim);
		bool add_sid(const Point& pt = Point::invalid);
		bool add_sid(const AirportsDb::Airport& arpt);
		bool add_star(const Point& pt = Point::invalid);
		bool add_star(const AirportsDb::Airport& arpt);
		void set_metric(void);
		void exclude_airspace(int baselevel, int toplevel, double maxarea = std::numeric_limits<double>::max());
		void exclude_regions(void);
		bool autoroute(FPlanRoute& route);
		const Graph& get_graph(void) const { return m_graph; }
		Graph& get_graph(void) { return m_graph; }

	protected:
		CFMUAutoroute& m_ar;
		Graph m_graph;
		Rect m_bbox;
		Graph::vertex_descriptor m_dep;
		Graph::vertex_descriptor m_dest;
		typedef std::vector<AirportsDb::Airport::VFRRoute> sidstar_t;
		sidstar_t m_sid;
		sidstar_t m_star;

		Graph::vertex_descriptor find_airport(const AirportsDb::Airport& arpt);

		void remove_sid(void);
		void remove_star(void);
	};

	class AirspaceCache;

	class DctCache {
	public:
		typedef unsigned long pointid_t;
		typedef TrafficFlowRestrictions::DctParameters::AltRange::set_t set_t;

		class PointID {
		public:
			PointID(const std::string& ident = "", const Point& pt = Point(), pointid_t id = 0) : m_ident(ident), m_point(pt), m_id(id) {}
			const std::string& get_ident(void) const { return m_ident; }
			const Point& get_point(void) const { return m_point; }
			pointid_t get_id(void) const { return m_id; }
			bool operator<(const PointID& x) const;

		protected:
			std::string m_ident;
			Point m_point;
			pointid_t m_id;
		};

		class DctAlt {
		public:
			DctAlt(pointid_t pt0 = 0, pointid_t pt1 = 0, const set_t& alt = set_t(), bool dirty = false) : m_alt(alt), m_point0(pt0), m_point1(pt1), m_dirty(dirty) {}
			const set_t& get_alt(void) const { return m_alt; }
		        set_t& get_alt(void) { return m_alt; }
			pointid_t get_point0(void) const { return m_point0; }
			pointid_t get_point1(void) const { return m_point1; }
			bool operator<(const DctAlt& x) const;
			bool is_dirty(void) const { return m_dirty; }
			void set_dirty(bool x = true) { m_dirty = true; }

		protected:
			set_t m_alt;
			pointid_t m_point0;
			pointid_t m_point1;
			bool m_dirty;
		};

		DctCache(const std::string& path = "", const std::string& revision = "1900-01-01T00:00:00");

		std::string get_property(const std::string& key);
		void set_property(const std::string& key, const std::string& value);

		void load(const Rect& bbox);
		void flush(void);

		pointid_t find_point(const std::string& ident, const Point& pt);
		const DctAlt *find_dct(pointid_t pt0, pointid_t pt1) const;
		void add_dct(pointid_t pt0, pointid_t pt1, const set_t& alt);

		unsigned int get_nrpoints(void) const { return m_points.size(); }
		unsigned int get_nrdcts(void) const { return m_dcts.size(); }

	protected:
		sqlite3x::sqlite3_connection m_db;
		typedef std::set<PointID> points_t;
		points_t m_points;
		typedef std::set<DctAlt> dcts_t;
		dcts_t m_dcts;
		pointid_t m_pointid;
		pointid_t m_pointidalloc;
	};

	typedef uint32_t lgraphairwayindex_t;
	static const lgraphairwayindex_t airwaymatchall = 0x0003FFFFU;
	static const lgraphairwayindex_t airwaymatchnone = airwaymatchall - 1U;
	static const lgraphairwayindex_t airwaymatchawy = airwaymatchnone - 1U;
	static const lgraphairwayindex_t airwaymatchdctawy = airwaymatchawy - 1U;
	static const lgraphairwayindex_t airwaymatchdctawysidstar = airwaymatchdctawy - 1U;
	static const lgraphairwayindex_t airwaymatchawysidstar = airwaymatchdctawysidstar - 1U;
	static const lgraphairwayindex_t airwaymatchsidstar = airwaymatchawysidstar - 1U;
	static const lgraphairwayindex_t airwaysid = airwaymatchsidstar - 1U;
	static const lgraphairwayindex_t airwaystar = airwaysid - 1U;
	static const lgraphairwayindex_t airwaydct = airwaystar - 1U;
	typedef std::set<std::string> lgraphairwayidents_t;

	class LVertex {
	public:
		LVertex(Engine::DbObject::ptr_t obj = Engine::DbObject::ptr_t());

		bool is_airport(void) const;
		bool is_navaid(void) const;
		bool is_intersection(void) const;
		bool is_mapelement(void) const;

		bool get(AirportsDb::Airport& el) const;
		bool get(NavaidsDb::Navaid& el) const;
		bool get(WaypointsDb::Waypoint& el) const;
		bool get(MapelementsDb::Mapelement& el) const;

		bool set(FPlanWaypoint& el) const;
		unsigned int insert(FPlanRoute& route, uint32_t nr = ~0U) const;

		Engine::DbObject::ptr_t get_object(void) const { return m_obj; }
		const std::string& get_ident(void) const { return m_ident; }
		bool is_ident_valid(void) const;
		const Point& get_coord(void) const { return m_coord; }

	protected:
		Engine::DbObject::ptr_t m_obj;
		Point m_coord;
		std::string m_ident;
	};

	class LEdge {
	public:
		static const float maxmetric;
		static const float invalidmetric;
		static const lgraphairwayindex_t identmask    = 0x0003FFFFU;
		static const lgraphairwayindex_t levelsmask   = 0x01FC0000U;
		static const lgraphairwayindex_t solutionmask = 0xFE000000U;

		LEdge(unsigned int levels = 0, lgraphairwayindex_t id = airwaydct, float dist = maxmetric, float truetrack = 0);
		LEdge(const LEdge& x);
		~LEdge();
		LEdge& operator=(const LEdge& x);
		void swap(LEdge& x);
		void resize(unsigned int levels);
		unsigned int get_levels(void) const { return (m_ident & levelsmask) >> 18; }
		void set_ident(lgraphairwayindex_t id) { m_ident ^= (m_ident ^ id) & identmask; }
		lgraphairwayindex_t get_ident(void) const { return m_ident & identmask; }
		bool is_dct(void) const { return get_ident() == airwaydct; }
		static bool is_match(lgraphairwayindex_t edgeid, lgraphairwayindex_t id);
		bool is_match(lgraphairwayindex_t id) const { return is_match(get_ident(), id); }
		void clear_solution(void) { m_ident |= solutionmask; }
		void set_solution(unsigned int pi) { m_ident ^= ((m_ident ^ (pi << 25)) & solutionmask); }
		unsigned int get_solution(void) const { return (m_ident & solutionmask) >> 25; }
		bool is_solution(void) const { return !!(solutionmask & ~m_ident); }
		bool is_solution(unsigned int pi) const { lgraphairwayindex_t x(pi); x <<= 25; return !((x ^ m_ident) & solutionmask) && (solutionmask & ~x); }
		void set_dist(float dist = maxmetric) { m_dist = dist; }
		float get_dist(void) const { return m_dist; }
		void set_truetrack(float tt = 0) { m_truetrack = tt; }
		float get_truetrack(void) const { return m_truetrack; }
		void set_metric(unsigned int pi, float metric = invalidmetric);
		float get_metric(unsigned int pi) const;
		void set_solution_metric(float metric = invalidmetric) { set_metric(get_solution(), metric); }
		float get_solution_metric(void) const { return get_metric(get_solution()); }
		void set_all_metric(float metric = invalidmetric);
		void clear_metric(unsigned int pi) { set_metric(pi, invalidmetric); }
		void clear_solution_metric(void) { clear_metric(get_solution()); }
		void clear_metric(void) { set_all_metric(invalidmetric); }
		void scale_metric(float scale = 1, float offset = 0);
		bool is_valid(unsigned int pi) const;
		bool is_valid(void) const;
		bool is_solution_valid(void) const { return is_solution() && is_valid(get_solution()); }

	protected:
	        lgraphairwayindex_t m_ident;

	public:
		float m_dist;

	protected:
		float m_truetrack;
		float *m_metric;
	};

	class LGraph : public boost::adjacency_list<boost::multisetS, boost::vecS, boost::bidirectionalS, LVertex, LEdge> {
	public:
		bool is_valid_connection(unsigned int piu, unsigned int piv, edge_descriptor e) const;
		bool is_valid_connection(vertex_descriptor u, unsigned int piu, vertex_descriptor v, unsigned int piv, edge_descriptor e) const;
		bool is_valid_connection_precheck(unsigned int piu, unsigned int piv, edge_descriptor e) const;
		bool is_valid_connection_precheck(vertex_descriptor u, unsigned int piu, vertex_descriptor v, unsigned int piv, edge_descriptor e) const;
		std::pair<edge_descriptor,bool> find_edge(vertex_descriptor u, vertex_descriptor v, lgraphairwayindex_t awy = airwaymatchall) const;
		
	};

	typedef std::multimap<std::string,LGraph::vertex_descriptor> LVerticesIdentMap;

	class LVertexLevel {
	public:
		LVertexLevel(LGraph::vertex_descriptor v = 0, unsigned int pi = 0) : m_vertex(v), m_perfindex(pi) {}
		LGraph::vertex_descriptor get_vertex(void) const { return m_vertex; }
		void set_vertex(LGraph::vertex_descriptor v) { m_vertex = v; }
		unsigned int get_perfindex(void) const { return m_perfindex; }
		void set_perfindex(unsigned int pi) { m_perfindex = pi; }
		bool operator<(const LVertexLevel& x) const;
		bool operator>(const LVertexLevel& x) const { return x < *this; }
		bool operator==(const LVertexLevel& x) const;
		bool operator!=(const LVertexLevel& x) const { return !(*this == x); }

	protected:
		LGraph::vertex_descriptor m_vertex;
		unsigned int m_perfindex;
	};

	class LVertexLevels {
	public:
		LVertexLevels(LGraph::vertex_descriptor v = 0, unsigned int pi0 = 0, unsigned int pi1 = 0) : m_vertex(v), m_perfindex0(pi0), m_perfindex1(pi1) {}
		LGraph::vertex_descriptor get_vertex(void) const { return m_vertex; }
		void set_vertex(LGraph::vertex_descriptor v) { m_vertex = v; }
		unsigned int get_perfindex0(void) const { return m_perfindex0; }
		unsigned int get_perfindex1(void) const { return m_perfindex1; }
		void set_perfindex0(unsigned int pi) { m_perfindex0 = pi; }
		void set_perfindex1(unsigned int pi) { m_perfindex1 = pi; }
		bool operator<(const LVertexLevels& x) const;
		bool operator>(const LVertexLevels& x) const { return x < *this; }
		bool operator==(const LVertexLevels& x) const;
		bool operator!=(const LVertexLevels& x) const { return !(*this == x); }

	protected:
		LGraph::vertex_descriptor m_vertex;
		unsigned int m_perfindex0;
		unsigned int m_perfindex1;
	};

        class CompareVertexLevelsVertexOnly {
        public:
                bool operator()(const LVertexLevels& a, const LVertexLevels& b);
        };

	class LVertexDescEdge {
	public:
		LVertexDescEdge(unsigned int levels, LGraph::vertex_descriptor v = 0, lgraphairwayindex_t id = airwaydct,
				float dist = LEdge::maxmetric, float truetrack = 0) : m_edge(levels, id, dist, truetrack), m_vertex(v) {}
		LGraph::vertex_descriptor get_vertex(void) const { return m_vertex; }
		void set_vertex(LGraph::vertex_descriptor v) { m_vertex = v; }
		LEdge& get_edge(void) const { return const_cast<LEdge&>(m_edge); }
		bool operator<(const LVertexDescEdge& x) const { return get_vertex() < x.get_vertex(); }
		bool operator>(const LVertexDescEdge& x) const { return x < *this; }
		bool operator==(const LVertexDescEdge& x) const { return get_vertex() == x.get_vertex(); }
		bool operator!=(const LVertexDescEdge& x) const { return !(*this == x); }

	protected:
		LEdge m_edge;
		LGraph::vertex_descriptor m_vertex;
	};

	class LVertexEdge : public LVertexLevel {
	public:
		// v should be the target of e
		LVertexEdge(const LVertexLevel& v = LVertexLevel(), lgraphairwayindex_t e = airwaymatchnone) : LVertexLevel(v), m_edge(e) {}
		lgraphairwayindex_t get_edge(void) const { return m_edge; }
		bool is_match(lgraphairwayindex_t id) const { return LEdge::is_match(get_edge(), id); }
		bool operator<(const LVertexEdge& x) const;
		bool operator>(const LVertexEdge& x) const { return x < *this; }

	protected:
		lgraphairwayindex_t m_edge;
	};

	class LRoute : public std::vector <LVertexEdge> {
	public:
		LRoute(void) : m_metric(std::numeric_limits<double>::quiet_NaN()) {}

		bool operator<(const LRoute& x) const;
		void set_metric(double metric) { m_metric = metric; }
		double get_metric(void) const { return m_metric; }
		bool is_inf(void) const { return std::isinf(get_metric()); }
		bool is_nan(void) const { return std::isnan(get_metric()); }
		bool is_repeated_nodes(const LGraph& g) const;
		bool is_existing(const LGraph& g) const;
		std::pair<LGraph::edge_descriptor,bool> get_edge(size_type i, const LGraph& g) const;
		static std::pair<LGraph::edge_descriptor,bool> get_edge(const LVertexEdge& u, const LVertexEdge& v, const LGraph& g);
		static std::pair<LGraph::edge_descriptor,bool> get_edge(LGraph::vertex_descriptor u, const LVertexEdge& v, const LGraph& g);

	protected:
		double m_metric;
	};

	class LGraphSolutionTreeNode;
	class LGraphSolutionTreeNode : public std::map<LVertexEdge,LGraphSolutionTreeNode> {
	};

	class LEdgeAdd {
	public:
		LEdgeAdd(LGraph::vertex_descriptor v0 = 0, LGraph::vertex_descriptor v1 = 0, double metric = std::numeric_limits<double>::max());
		LGraph::vertex_descriptor get_v0(void) const { return m_v0; }
		LGraph::vertex_descriptor get_v1(void) const { return m_v1; }
		double get_metric(void) const { return m_metric; }
		bool operator<(const LEdgeAdd& x) const;

	protected:
		LGraph::vertex_descriptor m_v0;
		LGraph::vertex_descriptor m_v1;
		double m_metric;
	};

	class LGraphAwyEdgeFilter {
	public:
		LGraphAwyEdgeFilter(void) : m_lgraph(0), m_airway(airwaymatchnone) {}
		LGraphAwyEdgeFilter(LGraph& g, lgraphairwayindex_t awy) : m_lgraph(&g), m_airway(awy) {}
		template <typename E> bool operator()(const E& e) const { return (*m_lgraph)[e].is_match(m_airway); }

	protected:
		LGraph *m_lgraph;
		lgraphairwayindex_t m_airway;
	};

	class LGraphSolutionEdgeFilter {
	public:
		LGraphSolutionEdgeFilter(void) : m_lgraph(0) {}
		LGraphSolutionEdgeFilter(LGraph& g) : m_lgraph(&g) {}
		template <typename E> bool operator()(const E& e) const { return (*m_lgraph)[e].is_solution(); }

	protected:
		LGraph *m_lgraph;
	};

	class DijkstraCore;
	class LevelEdgeWeightMap;
	friend double get(const LevelEdgeWeightMap& wm, LGraph::edge_descriptor e);

	class LGMandatoryPoint {
	public:
		LGMandatoryPoint(LGraph::vertex_descriptor v = 0, unsigned int pi0 = 0, unsigned int pi1 = 0,
				 lgraphairwayindex_t awy = airwaymatchnone) : m_v(v), m_pi0(pi0), m_pi1(pi1), m_airway(awy) {}
	        LGraph::vertex_descriptor get_v(void) const { return m_v; }
	        unsigned int get_pi0(void) const { return m_pi0; }
	        unsigned int get_pi1(void) const { return m_pi1; }
		lgraphairwayindex_t get_airway(void) const { return m_airway; }

	protected:
		LGraph::vertex_descriptor m_v;
		unsigned int m_pi0;
		unsigned int m_pi1;
		lgraphairwayindex_t m_airway;
	};

	class LGMandatorySequence : public std::vector<LGMandatoryPoint> {
	};

	class LGMandatoryAlternative : public std::vector<LGMandatorySequence> {
	public:
		LGMandatoryAlternative(const std::string& rule_id = "") : m_rule(rule_id) {}
		const std::string& get_rule(void) const { return m_rule; }
		void set_rule(const std::string& rule_id) { m_rule = rule_id; }

	protected:
		std::string m_rule;
	};

	class LGMandatory : public std::vector<LGMandatoryAlternative> {
	};

	class CFMUIntel {
	public:
		class ForbiddenPoint {
		public:
			ForbiddenPoint(const std::string& ident = "", const Point& pt = Point::invalid);
			const std::string& get_ident(void) const { return m_ident; }
			const Point& get_pt(void) const { return m_pt; }
			bool operator<(const ForbiddenPoint& x) const;
			bool is_valid(void) const;
			void save(sqlite3x::sqlite3_command& cmd) const;
			void load(sqlite3x::sqlite3_cursor& cursor);

		protected:
			std::string m_ident;
			Point m_pt;
		};

		class ForbiddenSegment {
		public:
			ForbiddenSegment(const std::string& fromident = "", const Point& frompt = Point::invalid, unsigned int fromalt = 0,
					 const std::string& toident = "", const Point& topt = Point::invalid, unsigned int toalt = 0,
					 lgraphairwayindex_t awy = airwaymatchnone, const std::string& awyident = "");
			const std::string& get_fromident(void) const { return m_fromident; }
			const std::string& get_toident(void) const { return m_toident; }
			const std::string& get_awyident(void) const { return m_awyident; }
			const Point& get_frompt(void) const { return m_frompt; }
			const Point& get_topt(void) const { return m_topt; }
			unsigned int get_fromalt(void) const { return m_fromalt; }
			unsigned int get_toalt(void) const { return m_toalt; }
		        lgraphairwayindex_t get_awyident(std::string& id) const;
			bool operator<(const ForbiddenSegment& x) const;
			bool is_valid(void) const;
			void save(sqlite3x::sqlite3_command& cmd) const;
			void load(sqlite3x::sqlite3_cursor& cursor);

		protected:
			std::string m_fromident;
			std::string m_toident;
			std::string m_awyident;
			Point m_frompt;
			Point m_topt;
			unsigned int m_fromalt;
			unsigned int m_toalt;
		};

		CFMUIntel();
		~CFMUIntel();
		void clear(void);
		void flush(void);
		void add(const ForbiddenPoint& p);
		void add(const ForbiddenSegment& s);
		void open(const std::string& maindir, const std::string& auxdir = "");
		void close(void);
		bool is_open(void) const;
		bool is_empty(void) const;
		void find(const Rect& bbox,
			  const sigc::slot<void,const ForbiddenPoint&>& ptfunc = sigc::slot<void,const ForbiddenPoint&>());
		void find(const Rect& bbox,
			  const sigc::slot<void,const ForbiddenSegment&>& segfunc = sigc::slot<void,const ForbiddenSegment&>());

	protected:
		sqlite3x::sqlite3_connection m_db;
		typedef std::set<ForbiddenPoint> forbiddenpoints_t;
		forbiddenpoints_t m_forbiddenpoints;
		typedef std::set<ForbiddenSegment> forbiddensegments_t;
		forbiddensegments_t m_forbiddensegments;
	};

	LGraph m_lgraph;
	typedef std::vector<std::string> lgraphairwaytable_t;
	lgraphairwaytable_t m_lgraphairwaytable;
	typedef std::map<std::string,lgraphairwayindex_t> lgraphairwayrevtable_t;
	lgraphairwayrevtable_t m_lgraphairwayrevtable;
	LVerticesIdentMap m_lverticesidentmap;
	typedef std::set<LVertexLevel> solutionvertices_t;
	solutionvertices_t m_solutionvertices;
	// Yen's k shortest loopless paths algorithm
	LGraphSolutionTreeNode m_solutiontree;
	typedef std::set<LRoute> lgraphsolutionpool_t;
	lgraphsolutionpool_t m_solutionpool;
	LGraph::vertex_descriptor m_vertexdep;
	LGraph::vertex_descriptor m_vertexdest;
	LGMandatory m_crossingmandatory;
	CFMUIntel m_cfmuintel;
	static const bool lgraphrestartedgescan = false;
	static const bool lgraphawyvertdct = false;

	void clearlgraph(void);
	lgraphairwayindex_t lgraphairway2index(const std::string& awy, bool create = false);
	const std::string& lgraphindex2airway(lgraphairwayindex_t idx, bool specialnames = false);
        LGraph::vertex_descriptor lgraphaddvertex(Engine::DbObject::ptr_t obj);
	typedef enum {
		setmetric_none,
		setmetric_dist,
		setmetric_metric
	} setmetric_t;
	void lgraphaddedge(LGraph& g, LEdge& ee, LGraph::vertex_descriptor u, LGraph::vertex_descriptor v, setmetric_t setmetric = setmetric_none, bool log = false);
	void lgraphaddedge(LEdge& ee, LGraph::vertex_descriptor u, LGraph::vertex_descriptor v, setmetric_t setmetric = setmetric_none, bool log = false) {
		lgraphaddedge(m_lgraph, ee, u, v, setmetric, log);
	}
	void lgraphkilledge(LGraph& g, LGraph::edge_descriptor e, unsigned int pi, bool log = false);
	void lgraphkillsolutionedge(LGraph& g, LGraph::edge_descriptor e, bool log = false) { lgraphkilledge(g, e, m_lgraph[e].get_solution(), log); }
	void lgraphkillalledges(LGraph& g, LGraph::edge_descriptor e, bool log = false);
	void lgraphkilledge(LGraph::edge_descriptor e, unsigned int pi, bool log = false) { lgraphkilledge(m_lgraph, e, pi, log); }
	void lgraphkillsolutionedge(LGraph::edge_descriptor e, bool log = false) { lgraphkilledge(e, m_lgraph[e].get_solution(), log); }
	void lgraphkillalledges(LGraph::edge_descriptor e, bool log = false) { lgraphkillalledges(m_lgraph, e, log); }
	LEdge lgraphbestedgemetric(LGraph::vertex_descriptor u, LGraph::vertex_descriptor v, lgraphairwayindex_t awy = airwaymatchall);
	bool lgraphkillemptyedges(void);
	bool lgraphload(const Rect& bbox);
	void lgraphkillinvalidsuper(void);
	static bool lgraphcheckintersect(const Point& pt0, const Point& pt1, const Rect& bbox);
	static bool lgraphcheckintersect(const Point& pt0, const Point& pt1, const Rect& bbox, const MultiPolygonHole& poly);
	void lgraphexcluderegions(void);
	void lgraphadddct(void);
	void lgraphkilllongdct(void);
	void lgraphedgemetric(void);
	bool lgraphaddsidstar(void);
	void lgraphapplycfmuintel(const Rect& bbox);
	void lgraphapplycfmuintelpt(const CFMUIntel::ForbiddenPoint& p);
	void lgraphapplycfmuintelseg(const CFMUIntel::ForbiddenSegment& s);
	void lgraphcrossing(void);
	static void lgraphedgemetric(double& cruisemetric, double& levelchgmetric, double& trknmi, unsigned int piu, unsigned int piv,
				     const LEdge& ee, const Performance& perf);
	void lgraphupdatemetric(LGraph& g, LRoute& r);
	void lgraphupdatemetric(LRoute& r) { lgraphupdatemetric(m_lgraph, r); }
	void lgraphclearcurrent(void);
	bool lgraphsetcurrent(const LRoute& r);
	void lgraphmodified(void);
	bool lgraphisinsolutiontree(const LRoute& r) const;
	bool lgraphinsertsolutiontree(const LRoute& r);
	std::string lgraphprint(LGraph& g, const LRoute& r);
	std::string lgraphprint(const LRoute& r) { return lgraphprint(m_lgraph, r); }
	LRoute lgraphdijkstra(LGraph& g, const LVertexLevel& u, const LVertexLevel& v, const LRoute& baseroute = LRoute(),
			      bool solutionedgeonly = false, LGMandatory mandatory = LGMandatory());
	LRoute lgraphdijkstra(const LVertexLevel& u, const LVertexLevel& v, const LRoute& baseroute = LRoute(),
			      bool solutionedgeonly = false, LGMandatory mandatory = LGMandatory()) {
		return lgraphdijkstra(m_lgraph, u, v, baseroute, solutionedgeonly, mandatory);
	}
	LRoute lgraphnextsolution(void);
	void lgraphroute(void);
	bool lgraphroutelocaltfr(void);
	static std::string lgraphlocaltfrseqstr(const TrafficFlowRestrictions::RuleResult::Alternative::Sequence& el);
	void lgraphloglocaltfr(const TrafficFlowRestrictions::RuleResult& rule, bool logvalidation = false);
	bool lgraphlocaltfrcheck(TrafficFlowRestrictions::Result& res);
	LGMandatory lgraphmandatoryroute(const TrafficFlowRestrictions::Result& res);
	void lgraphsubmitroute(void);
	bool lgraphtfrrulekillforbidden(LGraph& g, const TrafficFlowRestrictions::RuleResult::Alternative::Sequence& seq, bool log = true);
	bool lgraphtfrrulekillforbidden(const TrafficFlowRestrictions::RuleResult::Alternative::Sequence& seq, bool log = true) {
		return lgraphtfrrulekillforbidden(m_lgraph, seq, log);
	}
	IntervalSet<int32_t> lgraphcrossinggate(const TrafficFlowRestrictions::RuleResult::Alternative::Sequence& seq,
						const TrafficFlowRestrictions::RuleResult::Alternative::Sequence& cc);
	bool lgraphtfrruleresult(const TrafficFlowRestrictions::RuleResult& rule, const TrafficFlowRestrictions::Fpl& fpl);
	void lgraphtfrmessage(const TrafficFlowRestrictions::Message& msg, const TrafficFlowRestrictions::Fpl& fpl);
	std::string lgraphvertexname(const LGraph& g, LGraph::vertex_descriptor v);
	std::string lgraphvertexname(const LGraph& g, LGraph::vertex_descriptor v, unsigned int pi);
	std::string lgraphvertexname(const LGraph& g, LGraph::vertex_descriptor v, unsigned int pi0, unsigned int pi1);
	std::string lgraphvertexname(const LGraph& g, const LVertexLevel& v) { return lgraphvertexname(g, v.get_vertex(), v.get_perfindex()); }
	std::string lgraphvertexname(LGraph::vertex_descriptor v) { return lgraphvertexname(m_lgraph, v); }
	std::string lgraphvertexname(LGraph::vertex_descriptor v, unsigned int pi) { return lgraphvertexname(m_lgraph, v, pi); }
	std::string lgraphvertexname(LGraph::vertex_descriptor v, unsigned int pi0, unsigned int pi1) { return lgraphvertexname(m_lgraph, v, pi0, pi1); }
	std::string lgraphvertexname(const LVertexLevel& v) { return lgraphvertexname(v.get_vertex(), v.get_perfindex()); }
	void lgraphlogkillalledges(const LGraph& g, LGraph::edge_descriptor e);
	void lgraphlogkilledge(const LGraph& g, LGraph::edge_descriptor e, unsigned int pi);
	void lgraphlogkillsolutionedge(const LGraph& g, LGraph::edge_descriptor e) { lgraphlogkilledge(e, m_lgraph[e].get_solution()); }
	void lgraphlogkillalledges(LGraph::edge_descriptor e) { lgraphlogkillalledges(m_lgraph, e); }
	void lgraphlogkilledge(LGraph::edge_descriptor e, unsigned int pi) { lgraphlogkilledge(m_lgraph, e, pi); }
	void lgraphlogkillsolutionedge(LGraph::edge_descriptor e) { lgraphlogkillsolutionedge(m_lgraph, e); }
	void lgraphlogaddedge(const LGraph& g, LGraph::edge_descriptor e, unsigned int pi);
	void lgraphlogaddedge(LGraph::edge_descriptor e, unsigned int pi) { lgraphlogaddedge(m_lgraph, e, pi); }
	void lgraphlogrenameedge(const LGraph& g, LGraph::edge_descriptor e, unsigned int pi, lgraphairwayindex_t awyname);
	void lgraphlogrenameedge(LGraph::edge_descriptor e, unsigned int pi, lgraphairwayindex_t awyname) { lgraphlogrenameedge(m_lgraph, e, pi, awyname); }
	void lgraphlogscaleedge(const LGraph& g, LGraph::edge_descriptor e, unsigned int pi, double scale);
	void lgraphlogscaleedge(LGraph::edge_descriptor e, unsigned int pi, double scale) { lgraphlogscaleedge(m_lgraph, e, pi, scale); }
	bool lgraphdisconnectvertex(const std::string& ident, int lvlfrom = 0, int lvlto = 600, bool intel = false);
	bool lgraphdisconnectsolutionvertex(const std::string& ident);
	bool lgraphdisconnectsolutionedge(const std::string& id0, const std::string& id1, lgraphairwayindex_t awyname = airwaymatchall);
	bool lgraphscalesolutionvertex(const std::string& ident, double scale);
	bool lgraphscalesolutionedge(const std::string& id0, const std::string& id1, double scale);
	bool lgraphmodifyedge(const std::string& id0, const std::string& id1, lgraphairwayindex_t awyremove,
			      lgraphairwayindex_t awyadd = airwaymatchnone, int lvlfrom = 0, int lvlto = 600, bool bidir = false);
	bool lgraphmodifyawysegment(const std::string& id0, const std::string& id1, lgraphairwayindex_t awymatch,
				    lgraphairwayindex_t awyrename = airwaymatchnone, int lvlfrom = 0, int lvlto = 600, bool bidir = false, bool solutiononly = false);
	bool lgraphmodifysolutionsegment(const std::string& id0, const std::string& id1, lgraphairwayindex_t awymatch,
					 lgraphairwayindex_t awyrename = airwaymatchnone, int lvlfrom = 0, int lvlto = 600, bool matchbothid = true);
	bool lgraphchangeoutgoing(const std::string& id, lgraphairwayindex_t awyname);
	bool lgraphchangeincoming(const std::string& id, lgraphairwayindex_t awyname);
	bool lgraphedgetodct(lgraphairwayindex_t awy);
	bool lgraphkilloutgoing(const std::string& id);
	bool lgraphkillincoming(const std::string& id);
	bool lgraphkillsid(const std::string& id);
	bool lgraphkillstar(const std::string& id);
	bool lgraphkillclosesegments(const std::string& id0, const std::string& id1, const std::string& awyid,
				     lgraphairwayindex_t solutionmatch = airwaymatchdctawysidstar);
	bool lgraphkillclosesegments(const std::string& id0, const std::string& id1, lgraphairwayindex_t awy,
				     lgraphairwayindex_t solutionmatch = airwaymatchdctawysidstar);
	bool lgraphsolutiongroundclearance(void);
	template <typename T> static bool is_metric_invalid(T m) { return std::isnan(m) || std::isinf(m) || m == std::numeric_limits<T>::max(); }

	class GraphSerializeBase {
	public:
		GraphSerializeBase(void);

		const Rect& get_bbox(void) const { return m_bbox; }
		int16_t get_minlevel(void) const { return m_minlevel; }
		int16_t get_maxlevel(void) const { return m_maxlevel; }

		void set_bbox(const Rect& bbox) { m_bbox = bbox; }
		void set_minlevel(int16_t l) { m_minlevel = l; }
		void set_maxlevel(int16_t l) { m_maxlevel = l; }

		bool is_valid(void) const;

	protected:
		static const char signature[];
		Rect m_bbox;
		int16_t m_minlevel;
		int16_t m_maxlevel;
	};

	class GraphWrite : public GraphSerializeBase {
	public:
		GraphWrite(std::ostream& os);
		static bool is_save(void) { return true; }
		static bool is_load(void) { return false; }
		template <typename T> void io(const T& v);
		template <typename T> inline void ioint8(T v) { int8_t vv(v); io(vv); }
		template <typename T> inline void ioint16(T v) { int16_t vv(v); io(vv); }
		template <typename T> inline void ioint32(T v) { int32_t vv(v); io(vv); }
		template <typename T> inline void ioint64(T v) { int64_t vv(v); io(vv); }
		template <typename T> inline void iouint8(T v) { uint8_t vv(v); io(vv); }
		template <typename T> inline void iouint16(T v) { uint16_t vv(v); io(vv); }
		template <typename T> inline void iouint32(T v) { uint32_t vv(v); io(vv); }
		template <typename T> inline void iouint64(T v) { uint64_t vv(v); io(vv); }
		void header(void);

	protected:
		std::ostream& m_os;
	};

	class GraphRead : public GraphSerializeBase {
	public:
		GraphRead(std::istream& is);
		static bool is_save(void) { return false; }
		static bool is_load(void) { return true; }
		template <typename T> void io(T& v);
		template <typename T> inline void ioint8(T& v) { int8_t vv; io(vv); v = (T)vv; }
		template <typename T> inline void ioint16(T& v) { int16_t vv; io(vv); v = (T)vv; }
		template <typename T> inline void ioint32(T& v) { int32_t vv; io(vv); v = (T)vv; }
		template <typename T> inline void ioint64(T& v) { int64_t vv; io(vv); v = (T)vv; }
		template <typename T> inline void iouint8(T& v) { uint8_t vv; io(vv); v = (T)vv; }
		template <typename T> inline void iouint16(T& v) { uint16_t vv; io(vv); v = (T)vv; }
		template <typename T> inline void iouint32(T& v) { uint32_t vv; io(vv); v = (T)vv; }
		template <typename T> inline void iouint64(T& v) { uint64_t vv; io(vv); v = (T)vv; }
		void header(void);

	protected:
		std::istream& m_is;
	};

	class GraphFile;

	void lgraphsave(std::ostream& os, const Rect& bbox);
	void lgraphbinload(GraphRead& gr, const Rect& bbox);
	bool lgraphbinload(const Rect& bbox);
};

inline CFMUAutoroute::statusmask_t& operator|=(CFMUAutoroute::statusmask_t& a, CFMUAutoroute::statusmask_t b) { return a = (CFMUAutoroute::statusmask_t)(a | b); }
inline CFMUAutoroute::statusmask_t& operator&=(CFMUAutoroute::statusmask_t& a, CFMUAutoroute::statusmask_t b) { return a = (CFMUAutoroute::statusmask_t)(a & b); }
inline CFMUAutoroute::statusmask_t& operator^=(CFMUAutoroute::statusmask_t& a, CFMUAutoroute::statusmask_t b) { return a = (CFMUAutoroute::statusmask_t)(a ^ b); }
inline CFMUAutoroute::statusmask_t operator|(CFMUAutoroute::statusmask_t a, CFMUAutoroute::statusmask_t b) { a |= b; return a; }
inline CFMUAutoroute::statusmask_t operator&(CFMUAutoroute::statusmask_t a, CFMUAutoroute::statusmask_t b) { a &= b; return a; }
inline CFMUAutoroute::statusmask_t operator^(CFMUAutoroute::statusmask_t a, CFMUAutoroute::statusmask_t b) { a ^= b; return a; }

extern const std::string& to_str(CFMUAutoroute::opttarget_t t);
extern const std::string& to_str(CFMUAutoroute::log_t l);

#endif /* CFMUAUTOROUTE_H */
